"""Model for SDMX dataflows and related structures (like schemas).

``pysdmx`` is dataflow-centric, another area where ``pysdmx`` is
opinionated. As such, when retrieving information about a dataflow,
information typically provided via the data structure (and related
structures like concept schemes and codelists) is already provided
as part of the response.
"""

from collections import Counter, UserList
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Iterable, Literal, Optional, Sequence, Union

from msgspec import Struct

from pysdmx.errors import Invalid
from pysdmx.model.__base import (
    Agency,
    DataProvider,
    ItemReference,
    MaintainableArtefact,
)
from pysdmx.model.code import Codelist, Hierarchy
from pysdmx.model.concept import Concept, DataType, Facets


class Role(str, Enum):
    """The various roles a component can play."""

    DIMENSION = "D"
    """The component helps identifying data (e.g. primary key)."""
    MEASURE = "M"
    """The component holds a value we measure or collect."""
    ATTRIBUTE = "A"
    """The component provides descriptive information about the data."""


class ArrayBoundaries(Struct, frozen=True):
    """The minimum and maximum number of items in the SDMX array."""

    min_size: int = 0
    max_size: Optional[int] = None


class Component(Struct, frozen=True, omit_defaults=True):
    """A component of a dataset (aka **variable**), such the frequency.

    Concepts are used to **describe the relevant characteristics** of a
    statistical domain. For example, exchanges rates might be described with
    components such as the numerator currency, the denominator currency,
    the type of exchange rates, etc.

    Some of these components are expected to be useful **across statistical
    domains**. Examples of such components include the frequency, the
    observation status, the confidentiality, etc.

    When using components to describe the expected structure of a statistical
    domain, data stewards distinguish between the components that represent
    what is being captured (i.e. the **measures**), the components that help
    uniquely **identifying** the measures (i.e. the dimensions) and the
    components that provide additional **descriptive information** about the
    measures (i.e. the attributes). This is the **component role**. The role
    can be *D* (for Dimension), *A* (for Attribute) or *M* (for Measure).

    While dimensions and measures are typically mandatory, attributes may be
    either **mandatory or optional**. This is captured in the *required*
    property using a boolean value (*true* for mandatory components, *false*
    otherwise). This may vary with the statistical domain, i.e. a mandatory
    component within a particular domain may be optional in another.

    While the value of some attributes is expected to potentially vary with
    each *measurement* (aka observation or data point), some others must be
    unique across all observations sharing the same (sub)set of dimension
    values. This is captured in the *attachment_level* property, which can be
    one of: *D* (for Dataset), *O* (for Observation), any string identifying a
    component ID (FREQ) or comma-separated list of component IDs
    (FREQ,REF_AREA). The latter can be used to identify the dimension, group
    or series to which the attribute is attached. The attachment level of a
    component may vary with the statistical domain, i.e. a component attached
    to a series in a particular domain may be attached to, say, the dataset in
    another domain.

    The *codes* field indicates the expected (i.e. allowed) set of values a
    component can take within a particular domain. In addition to
    (or instead of) a set of codes, additional details about the expected
    format may be found in the *facets* and *dtype* fields.


    Attributes:
        id: A unique identifier for the component (e.g. FREQ).
        required: Whether the component must have a value.
        role: The role played by the component.
        local_dtype: The component's local data type (string, number, etc.).
        local_facets: Additional local details such as the component's minimum
            length.
        name: The component's name.
        description: Additional descriptive information about the component.
        local_codes: The expected local values for the component (e.g. currency
            codes).
        attachment_level: The attachement level (if role = A only).
            Attributes can be attached at different levels such as
            D (for dataset-level attributes), O (for observation-level
            attributes) or a combination of dimension IDs, separated by
            commas, for series- and group-level attributes).
        array_def: Any additional constraints for array types.
    """

    id: str
    required: bool
    role: Role
    concept: Union[Concept, ItemReference]
    local_dtype: Optional[DataType] = None
    local_facets: Optional[Facets] = None
    name: Optional[str] = None
    description: Optional[str] = None
    local_codes: Union[Codelist, Hierarchy, None] = None
    attachment_level: Optional[str] = None
    array_def: Optional[ArrayBoundaries] = None
    urn: Optional[str] = None

    def __post_init__(self) -> None:
        """Validate the component."""
        if self.role != Role.ATTRIBUTE and self.attachment_level is not None:
            raise Invalid(
                "Validation Error",
                (
                    "The attachment_level field is "
                    "only allowed for attribute components"
                ),
            )

    @property
    def dtype(self) -> DataType:
        """Returns the component data type.

        This will return the local data type (if any) or
        the data type of the referenced concept (if any).
        In case neither are set, the data type will default
        to string.

        Returns:
            The component data type (local, core or default).
        """
        if self.local_dtype:
            return self.local_dtype
        elif isinstance(self.concept, Concept) and self.concept.dtype:
            return self.concept.dtype
        else:
            return DataType.STRING

    @property
    def facets(self) -> Optional[Facets]:
        """Returns the component facets.

        This will return the local facets (if any) or
        the facets of the referenced concept (if any), or
        None in case neither are set.

        Returns:
            The component facets (local or core).
        """
        if self.local_facets:
            return self.local_facets
        elif isinstance(self.concept, Concept) and self.concept.facets:
            return self.concept.facets
        else:
            return None

    @property
    def enumeration(self) -> Union[Codelist, Hierarchy, None]:
        """Returns the list of valid codes for the component.

        This will return the local codes (if any) or
        the codes of the referenced concept (if any), or
        None in case neither are set.

        Returns:
            The component codes (local or core).
        """
        if self.local_codes:
            return self.local_codes
        elif isinstance(self.concept, Concept) and self.concept.codes:
            return self.concept.codes
        else:
            return None

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = []
        for k in self.__annotations__:
            v = self.__getattribute__(k)
            if v:
                if k == "concept":
                    out.append(f"{k}=({str(v)})")
                else:
                    out.append(f"{k}={str(v)}")
        return ", ".join(out)


class Components(UserList[Component]):
    """A collection of components describing the data."""

    def __init__(self, iterable: Iterable[Component]) -> None:
        """Create a new schema with the supplied components."""
        self.__validate_iterable(iterable, True)
        super().__init__(iterable)

    def __setitem__(self, index: Any, item: Any) -> None:
        """Add a component at the requested index."""
        self.__validate_comp(item)
        super().__setitem__(index, item)

    def __getitem__(self, i: Any) -> Any:
        """Return the component matching the supplied id or position.

        Args:
            i: The id of the component to be returned, or
                its position in the collection, or a slice.

        Returns:
            The component matching the supplied id or position, or
            multiple components in case slice is used.
        """
        if isinstance(i, (int, slice)):
            return super().__getitem__(i)
        else:
            out = list(filter(lambda item: item.id == i, self.data))
            if len(out) == 0:
                return None
            else:
                return out[0]

    def insert(self, i: int, item: Component) -> None:
        """Add a component at the requested index."""
        self.__validate_comp(item)
        super().insert(i, item)

    def append(self, item: Component) -> None:
        """Add a component to the existing list of components."""
        self.__validate_comp(item)
        super().append(item)

    def extend(self, other: Iterable[Component]) -> None:
        """Add the components to the existing list of components."""
        self.__validate_iterable(other, False)
        super().extend(other)

    @property
    def dimensions(self) -> Sequence[Component]:
        """Return the list of dimensions.

        Dimensions are components that contribute to the unique identification
        of a piece of data (aka an observation or data point). The
        combination of the values for all dimensions of an observation can
        therefore be seen as the observation's *primary key*.

        Returns:
            The list of dimensions
        """
        return [c for c in self.data if c.role == "D"]

    @property
    def attributes(self) -> Sequence[Component]:
        """Return the list of attributes.

        Attributes are components that provide descriptive information about
        some piece of data (aka an observation or data point).

        Returns:
            The list of attributes
        """
        return [c for c in self.data if c.role == "A"]

    @property
    def measures(self) -> Sequence[Component]:
        """Return the list of measures.

        Measures are components that hold the measured values.

        Returns:
            The list of measures
        """
        return [c for c in self.data if c.role == "M"]

    def __validate_iterable(
        self,
        coll: Iterable[Component],
        is_init: bool,
    ) -> None:
        if len(list(coll)) > 0:
            for fld in coll:
                self.__validate_comp(fld, is_init)
            counter = Counter([f.id for f in coll])
            dup = [i[0] for i in counter.items() if i[1] > 1]
            if len(dup) > 0:
                raise Invalid(
                    "Validation Error",
                    f"There are duplicates in the collection: {dup}",
                )

    def __validate_comp(self, fld: Component, is_init: bool = False) -> None:
        if not isinstance(fld, Component):
            raise Invalid(
                "Validation Error",
                f"Unexpected type. Expected Component but got: {type(fld)}",
            )
        if not is_init:
            ids = [f.id for f in self.data]
            if fld.id in ids:
                raise Invalid(
                    "Validation Error",
                    f"There is already a component with ID: {fld.id}",
                )


class DataflowInfo(Struct, frozen=True, omit_defaults=True):
    """Extended information about a dataflow.

    The information includes:

    - Some basic metadata about the dataflow (such as its ID and name).
    - Some useful metrics such as the number of observations.
    - The expected structure of data (i.e. the data schema), including
      the expected components, their types, etc.

    Attributes:
        id: The identifier of the dataflow (e.g. CBS).
        components: The data structure, i.e. the components, their types, etc.
        agency: The organization responsible for the data (e.g. BIS).
        name: The dataflow's name (e.g. Consolidated Banking Statistics).
        description: Additional descriptive information about the dataflow.
        version: The dataflow version.
        providers: The organizations providing the data.
        series_count: The number of series available in the dataflow.
        obs_count: The number of observations available in the dataflow.
        start_period: The oldest period for which data are available.
        end_period: The oldest period for which data are available.
        last_updated: When the dataflow was last updated.
        dsd_ref: The URN of the data structure used by the dataflow.
    """

    id: str
    components: Components
    agency: Agency
    name: Optional[str] = None
    description: Optional[str] = None
    version: str = "1.0"
    providers: Sequence[DataProvider] = ()
    series_count: Optional[int] = None
    obs_count: Optional[int] = None
    start_period: Optional[str] = None
    end_period: Optional[str] = None
    last_updated: Optional[datetime] = None
    dsd_ref: Optional[str] = None

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = []
        for k in self.__annotations__:
            v = self.__getattribute__(k)
            if v:
                out.append(f"{k}={v}")
        return ", ".join(out)


class Schema(Struct, frozen=True, omit_defaults=True):
    """The allowed content within a certain context.

    This is the equivalent to the result of a schema query in the
    SDMX-REST API.

    The response contains the list of allowed values for the
    selected context (one of datastructure, dataflow or
    provisionagreement), and is typically used for validation
    purposes.

    Attributes:
        context: The context for which the schema is provided.
            One of datastructure, dataflow or provisionagreement.
        agency: The agency maintaining the context (e.g. BIS).
        id: The ID of the context (e.g. BIS_MACRO).
        components: The list of components along with their
            allowed values, types, etc.
        version: The context version (e.g. 1.0)
        artefacts: The URNs of the artefacts used to generate the
            schema. This will typically include the URNs of data
            structures, codelists, concept schemes, content
            constraints, etc.
        generated: When the schema was generated. This is useful
            for metadata synchronization purposes. For example,
            if any of the artefacts listed under the artefacts
            property has been updated after the schema was
            generated, you might want to regenerate the schema.
    """

    context: Literal["datastructure", "dataflow", "provisionagreement"]
    agency: str
    id: str
    components: Components
    version: str = "1.0"
    artefacts: Sequence[str] = ()
    generated: datetime = datetime.now(timezone.utc)

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = []
        for k in self.__annotations__:
            v = self.__getattribute__(k)
            if v:
                out.append(f"{k}={v}")
        return ", ".join(out)

    @property
    def short_urn(self) -> str:
        """Returns a short URN for the schema."""
        SHORT_URN_MAPPING = {
            "datastructure": "DataStructure",
            "dataflow": "Dataflow",
            "provisionagreement": "ProvisionAgreement",
        }
        sdmx_type = SHORT_URN_MAPPING[self.context]
        return f"{sdmx_type}={self.agency}:{self.id}({self.version})"


class DataStructureDefinition(MaintainableArtefact, frozen=True, kw_only=True):
    """A collection of metadata concepts.

    Attributes:
        agency: The organization responsible for the data structure.
        annotations: The list of annotations attached to the data structure.
        id: The identifier of the data structure.
        description: descriptive information about the data structure.
        is_external_reference: Whether the data structure is an external
            reference.
        is_final: Whether the data structure is final.
        name: The data structure's name.
        service_url: The URL of the service providing the data structure.
        structure_url: The URL of the structure.
        uri: The URI of the data structure.
        urn: The URN of the data structure.
        valid_from: The date from which the data structure is valid.
        valid_to: The date until which the data structure is valid.
        version: The version of the data structure.
    """

    components: Components

    def __extract_artefacts(self) -> Sequence[str]:
        """Extract the artefacts used to generate the schema."""
        out = []
        for c in self.components:
            if c.local_codes:
                out.append(c.local_codes.urn)
            # Concept URNs
            if isinstance(c.concept, Concept):
                out.append(c.concept.urn)
                if c.concept.codes:
                    out.append(c.concept.codes.urn)
            else:
                urn_header = "urn:sdmx:org.sdmx.infomodel.conceptscheme."
                out.append(urn_header + str(c.concept))
        result = list({a for a in out if a})
        return result

    def to_schema(self) -> Schema:
        """Generates a Schema class from the DataStructureDefinition."""
        return Schema(
            context="datastructure",
            agency=(
                self.agency.id
                if isinstance(self.agency, Agency)
                else self.agency
            ),
            id=self.id,
            components=self.components,
            version=self.version,
            artefacts=self.__extract_artefacts(),
        )

    @property
    def short_urn(self) -> str:
        """Returns a short URN for the data structure."""
        return f"DataStructure={self.agency}:{self.id}({self.version})"


class Dataflow(
    MaintainableArtefact,
    frozen=True,
    omit_defaults=True,
    tag=True,
):
    """A flow of data that providers will provide."""

    structure: Optional[str] = None


class ProvisionAgreement(
    MaintainableArtefact, frozen=True, omit_defaults=True, kw_only=True
):
    """Link between a data provider and dataflow."""

    dataflow: str
    provider: str



"""Processes the input that comes into read_sdmx function."""

import csv
import os.path
from io import BytesIO, StringIO, TextIOWrapper
from json import JSONDecodeError, loads
from os import PathLike
from pathlib import Path
from typing import Tuple, Union

from httpx import get as httpx_get

from pysdmx.errors import Invalid, NotImplemented
from pysdmx.io.format import Format


def __remove_bom(input_string: str) -> str:
    return input_string.replace("\ufeff", "")


def __check_xml(input_str: str) -> bool:
    return input_str[:5] == "<?xml"


def __check_csv(input_str: str) -> bool:
    try:
        max_length = len(input_str) if len(input_str) < 1024 else 1024
        dialect = csv.Sniffer().sniff(input_str[:max_length])
        control_csv_format = (
            dialect.delimiter == "," and dialect.quotechar == '"'
        )
        # Check we can access the data and it is not empty
        if (
            len(input_str.splitlines()) > 1
            or input_str.splitlines()[0].count(",") > 1
        ) and control_csv_format:
            return True
    except Exception:
        return False
    return False


def __check_json(input_str: str) -> bool:
    try:
        loads(input_str)
        return True
    except JSONDecodeError:
        return False


def __get_sdmx_ml_flavour(input_str: str) -> Tuple[str, Format]:
    flavour_check = input_str[:1000].lower()
    if ":generic" in flavour_check:
        return input_str, Format.DATA_SDMX_ML_2_1_GEN
    if ":structurespecificdata" in flavour_check:
        return input_str, Format.DATA_SDMX_ML_2_1_STR
    if ":structure" in flavour_check:
        return input_str, Format.STRUCTURE_SDMX_ML_2_1
    if ":registryinterface" in flavour_check:
        return input_str, Format.REGISTRY_SDMX_ML_2_1
    if ":error" in flavour_check:
        return input_str, Format.ERROR_SDMX_ML_2_1
    raise Invalid("Validation Error", "Cannot parse input as SDMX-ML.")


def __get_sdmx_csv_flavour(input_str: str) -> Tuple[str, Format]:
    headers = csv.reader(StringIO(input_str)).__next__()
    if "DATAFLOW" in headers:
        return input_str, Format.DATA_SDMX_CSV_1_0_0
    elif "STRUCTURE" in headers and "STRUCTURE_ID" in headers:
        return input_str, Format.DATA_SDMX_CSV_2_0_0
    raise Invalid("Validation Error", "Cannot parse input as SDMX-CSV.")


def __check_sdmx_str(input_str: str) -> Tuple[str, Format]:
    """Attempts to infer the SDMX format of the input string."""
    if __check_xml(input_str):
        return __get_sdmx_ml_flavour(input_str)
    if __check_csv(input_str):
        return __get_sdmx_csv_flavour(input_str)
    if __check_json(input_str):
        raise NotImplemented("JSON formats reading are not supported yet")
    raise Invalid("Validation Error", "Cannot parse input as SDMX.")


def process_string_to_read(
    sdmx_document: Union[str, Path, BytesIO],
) -> Tuple[str, Format]:
    """Processes the input that comes into read_sdmx function.

    Automatically detects the format of the input. The input can be a file,
    URL, or string.

    Args:
        sdmx_document: Path to file, URL, or string.

    Returns:
        tuple: Tuple containing the parsed input and the format of the input.

    Raises:
        Invalid: If the input cannot be parsed as SDMX.
    """
    if isinstance(sdmx_document, str) and os.path.exists(sdmx_document):
        sdmx_document = Path(sdmx_document)
    # Read file as string
    if isinstance(sdmx_document, (Path, PathLike)):
        with open(
            sdmx_document, "r", encoding="utf-8-sig", errors="replace"
        ) as f:
            out_str = f.read()

    # Read from BytesIO
    elif isinstance(sdmx_document, BytesIO):
        text_wrap = TextIOWrapper(
            sdmx_document, encoding="utf-8", errors="replace"
        )
        out_str = text_wrap.read()

    elif isinstance(sdmx_document, str):
        if sdmx_document.startswith("http"):
            try:
                response = httpx_get(sdmx_document, timeout=60)
                if (
                    response.status_code != 200
                    and "<?xml" not in response.text
                ):
                    raise Exception("Invalid URL, no SDMX Error found")
                out_str = response.text
            except Exception:
                raise Invalid(
                    "Validation Error",
                    f"Cannot retrieve a SDMX Message "
                    f"from URL: {sdmx_document}.",
                ) from None
        else:
            out_str = sdmx_document
    else:
        raise Invalid(
            "Validation Error",
            f"Cannot parse input of type {type(sdmx_document)}.",
        )

    out_str = __remove_bom(out_str)

    return __check_sdmx_str(out_str)



"""SDMX 2.1 XML Header reader module."""

import warnings
from typing import Any, Dict, Optional, Union

from pysdmx.io.xml.sdmx21.__tokens import (
    AGENCY_ID,
    DATASET_ACTION,
    DATASET_ID,
    DFW,
    DIM_OBS,
    DSD,
    GENERIC,
    HEADER,
    HEADER_ID,
    ID,
    NAME,
    PREPARED,
    PROV_AGREMENT,
    RECEIVER,
    REF,
    SENDER,
    SOURCE,
    STR_SPE,
    STR_USAGE,
    STRUCTURE,
    TEST,
    URN,
    VERSION,
)
from pysdmx.io.xml.sdmx21.reader.__parse_xml import parse_xml
from pysdmx.model import Organisation, Reference
from pysdmx.model.message import Header
from pysdmx.util import parse_urn


def __parse_sender_receiver(
    sender_receiver: Union[Dict[str, Any], None],
) -> Union[Organisation, None]:
    """Parses the sender or receiver of the SDMX message."""
    if sender_receiver is None:
        return None
    names = sender_receiver.get(NAME, None)
    selected_name = None

    if names is not None:
        if not isinstance(names, list):
            names = [names]

        if isinstance(names[0], str):
            selected_name = names[0]

        else:
            selected_name = next(
                (
                    name.get("#text")
                    for name in names
                    if name.get("lang", "en") == "en"
                ),
                names[0].get("#text"),
            )

    expected_keys = {NAME, ID}
    unexpected_keys = set(sender_receiver.keys()) - expected_keys
    unexpected_keys = {x for x in unexpected_keys if "xml" not in x.lower()}
    if unexpected_keys:
        warnings.warn(
            f"The following attributes will be lost: "
            f"{', '.join(unexpected_keys)}",
            UserWarning,
            stacklevel=2,
        )

    id_ = sender_receiver.get(ID, "ZZZ")

    organisation = Organisation(
        id=id_,
        name=selected_name,
    )

    return organisation


def __parse_structure(
    structure: Union[Dict[str, Any], None],
) -> Union[Dict[str, str], None]:
    """Parses the structure of the SDMX header."""
    if structure is None:
        return None

    dim_at_obs = structure.get(DIM_OBS, "AllDimensions")

    if STRUCTURE in structure:
        structure_info = structure[STRUCTURE]
        sdmx_type = DSD
    elif STR_USAGE in structure:
        structure_info = structure[STR_USAGE]
        sdmx_type = DFW
    else:
        structure_info = structure[PROV_AGREMENT]
        sdmx_type = PROV_AGREMENT

    if REF in structure_info:
        reference = structure_info[REF]
        agency_id = reference[AGENCY_ID]
        structure_id = reference[ID]
        version = reference[VERSION]
        ref_obj = Reference(
            sdmx_type=sdmx_type,
            agency=agency_id,
            id=structure_id,
            version=version,
        )
    else:
        ref_obj = parse_urn(structure_info[URN])
    return {str(ref_obj): dim_at_obs}


def __parse_header(header: Dict[str, Any]) -> Header:
    """Parses the header of the SDMX message.

    Args:
        header: Dictionary with the header information.

    Returns:
        Header: The header of the SDMX message.

    """
    dict_header = {
        "id": header.get(HEADER_ID),
        "test": header.get(TEST),
        "prepared": header.get(PREPARED),
        "sender": __parse_sender_receiver(header.get(SENDER)),
        "receiver": __parse_sender_receiver(header.get(RECEIVER)),
        "source": header.get(SOURCE),
        "dataset_action": header.get(DATASET_ACTION),
        "structure": __parse_structure(header.get(STRUCTURE)),
        "dataset_id": header.get(DATASET_ID),
    }

    return Header(**dict_header)  # type: ignore[arg-type]


def read(
    input_str: str,
    validate: bool = True,
) -> Optional[Header]:
    """Reads and retrieves the header of the SDMX message.

    Args:
        input_str: The input string to be parsed.
        validate: If True, the SDMX-ML data will be validated against the XSD.

    Returns:
        The header of the SDMX message.
    """
    dict_info = parse_xml(input_str, validate)
    possible_keys = [STR_SPE, GENERIC, STRUCTURE]
    selected_key = next((key for key in possible_keys if key in dict_info))
    if HEADER not in dict_info[selected_key]:
        return None
    header = dict_info[selected_key][HEADER]
    return __parse_header(header)



"""Collection of SDMX-JSON schemas for generic structure messages."""

from datetime import datetime
from typing import Dict, Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.category import (
    JsonCategorisation,
    JsonCategoryScheme,
)
from pysdmx.io.json.sdmxjson2.messages.code import (
    JsonCodelist,
    JsonHierarchy,
    JsonHierarchyAssociation,
    JsonValuelist,
)
from pysdmx.io.json.sdmxjson2.messages.concept import JsonConceptScheme
from pysdmx.io.json.sdmxjson2.messages.constraint import JsonDataConstraint
from pysdmx.io.json.sdmxjson2.messages.dataflow import JsonDataflow
from pysdmx.io.json.sdmxjson2.messages.dsd import JsonDataStructure
from pysdmx.io.json.sdmxjson2.messages.map import (
    JsonRepresentationMap,
    JsonStructureMap,
)
from pysdmx.io.json.sdmxjson2.messages.org import (
    JsonAgencyScheme,
    JsonDataProviderScheme,
)
from pysdmx.io.json.sdmxjson2.messages.pa import JsonProvisionAgreement
from pysdmx.model import Organisation


class Header(Struct, frozen=True):
    """The message header."""

    id: str
    prepared: datetime
    sender: Organisation
    test: bool = False
    content_languages: Sequence[str] = ()
    name: Optional[str] = None
    names: Optional[Dict[str, str]] = None
    receivers: Optional[Organisation] = None


class Structures(Struct, frozen=True):
    """The allowed strutures."""

    data_structures: Sequence[JsonDataStructure] = ()
    category_schemes: Sequence[JsonCategoryScheme] = ()
    concept_schemes: Sequence[JsonConceptScheme] = ()
    codelists: Sequence[JsonCodelist] = ()
    value_lists: Sequence[JsonValuelist] = ()
    hierarchies: Sequence[JsonHierarchy] = ()
    hierarchy_associations: Sequence[JsonHierarchyAssociation] = ()
    agency_schemes: Sequence[JsonAgencyScheme] = ()
    data_provider_schemes: Sequence[JsonDataProviderScheme] = ()
    dataflows: Sequence[JsonDataflow] = ()
    provision_agreements: Sequence[JsonProvisionAgreement] = ()
    structure_maps: Sequence[JsonStructureMap] = ()
    representation_maps: Sequence[JsonRepresentationMap] = ()
    categorisations: Sequence[JsonCategorisation] = ()
    dataConstraints: Sequence[JsonDataConstraint] = ()


class StructureMessage(Struct, frozen=True):
    """A generic SDMX-JSON 2.0 Structure message."""

    meta: Header
    data: Structures



"""Collection of SDMX-JSON schemas for provision agreements."""

from datetime import datetime
from typing import Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import JsonAnnotation
from pysdmx.model import ProvisionAgreement


class JsonProvisionAgreement(
    Struct, frozen=True, rename={"agency": "agencyID"}
):
    """SDMX-JSON payload for a provision agreement."""

    id: str
    name: str
    agency: str
    dataflow: str
    dataProvider: str
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None

    def to_model(self) -> ProvisionAgreement:
        """Converts a FusionPA to a standard provision agreement."""
        return ProvisionAgreement(
            id=self.id,
            agency=self.agency,
            name=self.name,
            description=self.description,
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            dataflow=self.dataflow,
            provider=self.dataProvider,
        )


class JsonProvisionAgreements(Struct, frozen=True):
    """SDMX-JSON payload for provision agreements."""

    provisionAgreements: Sequence[JsonProvisionAgreement]

    def to_model(self) -> Sequence[ProvisionAgreement]:
        """Returns the requested provision agreements."""
        return [pa.to_model() for pa in self.provisionAgreements]


class JsonProvisionAgreementsMessage(Struct, frozen=True):
    """SDMX-JSON payload for /provisionagreement queries."""

    data: JsonProvisionAgreements

    def to_model(self) -> Sequence[ProvisionAgreement]:
        """Returns the requested provision agreements."""
        return self.data.to_model()



"""Collection of Fusion-JSON schemas for provision agreements."""

from typing import Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.fusion.messages.core import FusionString
from pysdmx.model import ProvisionAgreement as PA


class FusionProvisionAgreement(
    Struct, frozen=True, rename={"agency": "agencyId"}
):
    """Fusion-JSON payload for a provision agreement."""

    id: str
    names: Sequence[FusionString]
    agency: str
    structureUsage: str
    dataproviderRef: str
    descriptions: Optional[Sequence[FusionString]] = None
    version: str = "1.0"

    def to_model(self) -> PA:
        """Converts a JsonPA to a standard provision agreement."""
        description = self.descriptions[0].value if self.descriptions else None
        return PA(
            id=self.id,
            name=self.names[0].value,
            agency=self.agency,
            description=description,
            version=self.version,
            provider=self.dataproviderRef,
            dataflow=self.structureUsage,
        )


class FusionProvisionAgreementMessage(Struct, frozen=True):
    """Fusion-JSON payload for /provisionagreement queries."""

    ProvisionAgreement: Sequence[FusionProvisionAgreement]

    def to_model(self) -> Sequence[PA]:
        """Returns the requested provision agreements."""
        return [c.to_model() for c in self.ProvisionAgreement]



from parsy import (  # type: ignore[import-untyped]
    from_enum,
    match_item,
    regex,
    seq,
    string,
)

from pysdmx.api.dc.query._model import Operator
from pysdmx.api.dc.query._parsing_model import (
    _CoreOperator,
    _Filter,
    _InOperator,
    _RangeOperator,
)
from pysdmx.api.dc.query._parsing_util import (
    _field,
    _in_vals,
    _pad,
    _to_response,
    _value,
)


def __up(input: str) -> str:
    return input.upper()


__lparen = match_item("(")
__rparen = match_item(")")

__co_op = from_enum(_CoreOperator, transform=__up)
__in_op = regex(r"(?i:NOT)?\s*(?i:IN){1}").map(__up).map(_InOperator)
__ra_op = regex(r"(?i:NOT)?\s*(?i:BETWEEN){1}").map(__up).map(_RangeOperator)


__between_vals = _value.sep_by(
    _pad + string("AND", transform=__up) + _pad,
    min=2,
    max=2,
)

__core_parser = seq(
    field=_pad >> _field << _pad,
    operator=__co_op.map(lambda o: Operator[o.name]),
    value=_pad >> _value << _pad,
).combine_dict(_Filter)

__in_parser = seq(
    field=_pad >> _field << _pad,
    operator=__in_op.map(lambda o: Operator[o.name]),
    value=_pad >> __lparen >> _in_vals << __rparen << _pad,
).combine_dict(_Filter)

__range_parser = seq(
    field=_pad >> _field << _pad,
    operator=__ra_op.map(lambda o: Operator[o.name]),
    value=_pad >> __between_vals << _pad,
).combine_dict(_Filter)


__single_parser = __core_parser | __in_parser | __range_parser
sql_parser = __single_parser.sep_by(
    _pad + string("AND", transform=__up) + _pad, min=1
).map(_to_response)

__all__ = ["sql_parser"]



"""Build SDMX-REST schema queries."""

from datetime import datetime
from enum import Enum
from typing import Optional

import msgspec
from msgspec.json import Decoder

from pysdmx.api.qb.structure import _V2_0_ADDED, StructureType
from pysdmx.api.qb.util import REST_ALL, REST_LATEST, ApiVersion, CoreQuery
from pysdmx.errors import Invalid
from pysdmx.io.format import SchemaFormat


class SchemaContext(Enum):
    """The context for which a schema must be generated."""

    DATA_STRUCTURE = "datastructure"
    METADATA_STRUCTURE = "metadatastructure"
    DATAFLOW = "dataflow"
    METADATA_FLOW = "metadataflow"
    PROVISION_AGREEMENT = "provisionagreement"
    METADATA_PROVISION_AGREEMENT = "metadataprovisionagreement"


class SchemaQuery(CoreQuery, frozen=True, omit_defaults=True):
    """A schema query.

    Schema queries allow retrieving the definition of data validity for a
    certain context. The service must take into account the constraints
    that apply within that context (e.g. dataflow).

    Attributes:
        context: The context for which a schema must be generated. This
            determines the constraints that will be taken into
            consideration.
        agency_id: The agency maintaining the context to be considered.
        resource_id: The id of the context to be considered.
        version: The version of the context to be considered.
        obs_dimension: The ID of the dimension at the observation level.
        explicit: For cross-sectional data validation, indicates whether
            observations are strongly typed.
        deletion: Whether the schema will be used to validate deletion
            messages.
        as_of: Retrieve the schema as it was at the specified point
            in time (aka time travel).
    """

    context: SchemaContext
    agency_id: str
    resource_id: str
    version: str = REST_LATEST
    obs_dimension: Optional[str] = None
    explicit: bool = False
    deletion: bool = False
    as_of: Optional[datetime] = None

    def __check_context(self, version: ApiVersion) -> None:
        ct = StructureType(self.context.value)
        if version < ApiVersion.V2_0_0 and ct in _V2_0_ADDED:
            raise Invalid(
                "Validation Error",
                f"{self.context} not allowed in {version.value}.",
            )

    def __check_version(self) -> None:
        if self.version == REST_ALL:
            raise Invalid(
                "Validation Error",
                "Retrieving schemas for all versions is not allowed.",
            )

    def __check_explicit(self, version: ApiVersion) -> None:
        if self.explicit and version >= ApiVersion.V2_0_0:
            raise Invalid(
                "Validation Error",
                f"Explicit parameter is not supported in {version.value}.",
            )

    def __check_as_of(self, version: ApiVersion) -> None:
        if self.as_of and version < ApiVersion.V2_2_0:
            raise Invalid(
                "Validation Error",
                f"as_of not supported in {version.value}.",
            )

    def __check_deletion(self, version: ApiVersion) -> None:
        if self.deletion and version < ApiVersion.V2_2_0:
            raise Invalid(
                "Validation Error",
                f"deletion parameter is not supported in {version.value}.",
            )

    def _validate_query(self, version: ApiVersion) -> None:
        self.validate()
        self.__check_context(version)
        self.__check_version()
        self.__check_explicit(version)
        self.__check_as_of(version)
        self.__check_deletion(version)

    def _create_full_query(self, ver: ApiVersion) -> str:
        u = (
            f"/schema/{self.context.value}/"
            f"{self.agency_id}/{self.resource_id}"
        )
        u += f"/{self._to_kw(self.version, ver)}"
        if (
            self.obs_dimension
            or self.as_of
            or ver < ApiVersion.V2_0_0
            or ver >= ApiVersion.V2_2_0
        ):
            u += "?"
        if self.obs_dimension:
            u += f"dimensionAtObservation={self.obs_dimension}"
        if ver < ApiVersion.V2_0_0:
            if self.obs_dimension:
                u += "&"
            u += f"explicit={str(self.explicit).lower()}"
        if self.as_of:
            if self.obs_dimension:
                u += "&"
            u += f'asOf={self.as_of.isoformat("T", "seconds")}'
        if ver >= ApiVersion.V2_2_0:
            if self.obs_dimension or self.as_of:
                u += "&"
            u += f"deletion={str(self.deletion).lower()}"
        return u

    def _create_short_query(self, ver: ApiVersion) -> str:
        u = (
            f"/schema/{self.context.value}/"
            f"{self.agency_id}/{self.resource_id}"
        )
        if self.version != REST_LATEST:
            u += f"/{self._to_kw(self.version, ver)}"
        if self.obs_dimension or self.explicit or self.as_of or self.deletion:
            u += "?"
        if self.obs_dimension:
            u += f"dimensionAtObservation={self.obs_dimension}"
        if self.explicit:
            if self.obs_dimension:
                u += "&"
            u += f"explicit={str(self.explicit).lower()}"
        if self.as_of:
            if self.obs_dimension:
                u += "&"
            u += f'asOf={self.as_of.isoformat("T", "seconds")}'
        if self.deletion:
            if self.obs_dimension or self.as_of:
                u += "&"
            u += f"deletion={str(self.deletion).lower()}"
        return u

    def _get_decoder(self) -> Decoder:  # type: ignore[type-arg]
        return _decoder


_decoder = msgspec.json.Decoder(SchemaQuery)

__all__ = ["SchemaContext", "SchemaFormat", "SchemaQuery"]



"""Collection of Fusion-JSON schemas for VTL artefacts."""

from datetime import datetime
from typing import Literal, Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.fusion.messages.core import FusionString
from pysdmx.model.vtl import (
    CustomType,
    CustomTypeScheme,
    FromVtlMapping,
    NamePersonalisation,
    NamePersonalisationScheme,
    Ruleset,
    RulesetScheme,
    ToVtlMapping,
    Transformation,
    UserDefinedOperator,
    UserDefinedOperatorScheme,
    VtlCodelistMapping,
    VtlConceptMapping,
    VtlDataflowMapping,
    VtlMapping,
    VtlMappingScheme,
)
from pysdmx.model.vtl import (
    TransformationScheme as TS,
)


class FusionCustomType(Struct, frozen=True):
    """Fusion-JSON payload for custom types."""

    id: str
    vtlScalarType: str
    dataType: str
    names: Sequence[FusionString] = ()
    vtlLiteralFormat: Optional[str] = None
    outputFormat: Optional[str] = None
    nullValue: Optional[str] = None
    descriptions: Sequence[FusionString] = ()

    def to_model(self) -> CustomType:
        """Converts deserialized class to pysdmx model class."""
        return CustomType(
            id=self.id,
            name=self.names[0].value,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            data_type=self.dataType,
            null_value=self.nullValue,
            output_format=self.outputFormat,
            vtl_literal_format=self.vtlLiteralFormat,
            vtl_scalar_type=self.vtlScalarType,
        )


class FusionCustomTypeScheme(
    Struct, frozen=True, rename={"agency": "agencyId"}
):
    """Fusion-JSON payload for custom type schemes."""

    id: str
    agency: str
    vtlVersion: str
    names: Sequence[FusionString] = ()
    items: Sequence[FusionCustomType] = ()
    descriptions: Sequence[FusionString] = ()
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    isPartial: bool = False

    def to_model(self) -> CustomTypeScheme:
        """Converts deserialized class to pysdmx model class."""
        items = [i.to_model() for i in self.items]
        return CustomTypeScheme(
            self.id,
            name=self.names[0].value,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            is_external_reference=self.isExternalReference,
            agency=self.agency,
            items=items,
            is_partial=self.isPartial,
            vtl_version=self.vtlVersion,
        )


class FusionNamePersonalisation(Struct, frozen=True):
    """Fusion-JSON payload for name personalisations."""

    id: str
    vtlDefaultName: str
    personalisedName: str
    vtlArtefact: str
    names: Sequence[FusionString] = ()
    descriptions: Sequence[FusionString] = ()

    def to_model(self) -> NamePersonalisation:
        """Converts deserialized class to pysdmx model class."""
        return NamePersonalisation(
            self.id,
            name=self.names[0].value,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            vtl_default_name=self.vtlDefaultName,
            personalised_name=self.personalisedName,
            vtl_artefact=self.vtlArtefact,
        )


class FusionNamePersonalisationScheme(
    Struct, frozen=True, rename={"agency": "agencyId"}
):
    """Fusion-JSON payload for name personalisation schemes."""

    id: str
    agency: str
    vtlVersion: str
    names: Sequence[FusionString] = ()
    items: Sequence[FusionNamePersonalisation] = ()
    descriptions: Sequence[FusionString] = ()
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    isPartial: bool = False

    def to_model(self) -> NamePersonalisationScheme:
        """Converts deserialized class to pysdmx model class."""
        items = [i.to_model() for i in self.items]
        return NamePersonalisationScheme(
            self.id,
            name=self.names[0].value,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            is_external_reference=self.isExternalReference,
            agency=self.agency,
            items=items,
            is_partial=self.isPartial,
            vtl_version=self.vtlVersion,
        )


class FusionUserDefinedOperator(Struct, frozen=True):
    """Fusion-JSON payload for user defined operator."""

    id: str
    operatorDefinition: str
    names: Sequence[FusionString] = ()
    descriptions: Sequence[FusionString] = ()

    def to_model(self) -> UserDefinedOperator:
        """Converts deserialized class to pysdmx model class."""
        return UserDefinedOperator(
            self.id,
            name=self.names[0].value,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            operator_definition=self.operatorDefinition,
        )


class FusionUserDefinedOperatorScheme(
    Struct, frozen=True, rename={"agency": "agencyId"}
):
    """Fusion-JSON payload for user defined operator schemes."""

    id: str
    agency: str
    vtlVersion: str
    names: Sequence[FusionString] = ()
    items: Sequence[FusionUserDefinedOperator] = ()
    vtlMappingScheme: Optional[str] = None
    rulesetSchemes: Sequence[str] = ()
    descriptions: Sequence[FusionString] = ()
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    isPartial: bool = False

    def to_model(self) -> UserDefinedOperatorScheme:
        """Converts deserialized class to pysdmx model class."""
        items = [i.to_model() for i in self.items]
        return UserDefinedOperatorScheme(
            self.id,
            name=self.names[0].value,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            is_external_reference=self.isExternalReference,
            agency=self.agency,
            items=items,
            is_partial=self.isPartial,
            vtl_version=self.vtlVersion,
            vtl_mapping_scheme=self.vtlMappingScheme,
            ruleset_schemes=self.rulesetSchemes,
        )


class FusionRuleset(Struct, frozen=True):
    """Fusion-JSON payload for rulesets."""

    id: str
    rulesetDefinition: str
    rulesetType: Literal["datapoint", "hierarchical"]
    rulesetScope: Literal["variable", "valuedomain"]
    names: Sequence[FusionString] = ()
    descriptions: Sequence[FusionString] = ()

    def to_model(self) -> Ruleset:
        """Converts deserialized class to pysdmx model class."""
        return Ruleset(
            id=self.id,
            name=self.names[0].value,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            ruleset_definition=self.rulesetDefinition,
            ruleset_type=self.rulesetType,
            ruleset_scope=self.rulesetScope,
        )


class FusionRulesetScheme(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for ruleset schemes."""

    id: str
    agency: str
    vtlVersion: str
    names: Sequence[FusionString] = ()
    items: Sequence[FusionRuleset] = ()
    vtlMappingScheme: Optional[str] = None
    descriptions: Sequence[FusionString] = ()
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    isPartial: bool = False

    def to_model(self) -> RulesetScheme:
        """Converts deserialized class to pysdmx model class."""
        items = [i.to_model() for i in self.items]
        return RulesetScheme(
            self.id,
            name=self.names[0].value,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            is_external_reference=self.isExternalReference,
            agency=self.agency,
            items=items,
            is_partial=self.isPartial,
            vtl_version=self.vtlVersion,
            vtl_mapping_scheme=self.vtlMappingScheme,
        )


class FusionVtlMapping(Struct, frozen=True):
    """Fusion-JSON payload for VTL mappings."""

    id: str
    alias: str
    mapped: str
    names: Sequence[FusionString] = ()
    descriptions: Sequence[FusionString] = ()
    toVtlMethod: Optional[str] = None
    toVtlSubSpace: Sequence[str] = ()
    fromVtlMethod: Optional[str] = None
    fromVtlSuperSpace: Sequence[str] = ()

    def to_model(self) -> VtlMapping:
        """Converts deserialized class to pysdmx model class."""
        if "Codelist=" in self.mapped:
            return VtlCodelistMapping(
                self.id,
                name=self.names[0].value,
                description=(
                    self.descriptions[0].value if self.descriptions else None
                ),
                codelist=self.mapped,
                codelist_alias=self.alias,
            )
        elif "Concept" in self.mapped:
            return VtlConceptMapping(
                self.id,
                name=self.names[0].value,
                description=(
                    self.descriptions[0].value if self.descriptions else None
                ),
                concept=self.mapped,
                concept_alias=self.alias,
            )
        else:
            return VtlDataflowMapping(
                self.id,
                name=self.names[0].value,
                description=(
                    self.descriptions[0].value if self.descriptions else None
                ),
                dataflow=self.mapped,
                dataflow_alias=self.alias,
                from_vtl_mapping_method=(
                    FromVtlMapping(self.fromVtlSuperSpace, self.fromVtlMethod)
                    if self.fromVtlMethod
                    else None
                ),
                to_vtl_mapping_method=(
                    ToVtlMapping(self.toVtlSubSpace, self.toVtlMethod)
                    if self.toVtlMethod
                    else None
                ),
            )


class FusionVtlMappingScheme(
    Struct, frozen=True, rename={"agency": "agencyId"}
):
    """Fusion-JSON payload for VTL mapping schemes."""

    id: str
    agency: str
    names: Sequence[FusionString] = ()
    items: Sequence[FusionVtlMapping] = ()
    descriptions: Sequence[FusionString] = ()
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    isPartial: bool = False

    def to_model(self) -> VtlMappingScheme:
        """Converts deserialized class to pysdmx model class."""
        items = [i.to_model() for i in self.items]
        return VtlMappingScheme(
            self.id,
            name=self.names[0].value,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            is_external_reference=self.isExternalReference,
            agency=self.agency,
            items=items,
            is_partial=self.isPartial,
        )


class FusionTransformation(Struct, frozen=True):
    """Fusion-JSON payload for VTL transformations."""

    id: str
    expression: str
    result: str
    isPersistent: bool
    names: Sequence[FusionString] = ()
    descriptions: Sequence[FusionString] = ()

    def to_model(self) -> Transformation:
        """Converts deserialized class to pysdmx model class."""
        return Transformation(
            self.id,
            name=self.names[0].value,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            expression=self.expression,
            is_persistent=self.isPersistent,
            result=self.result,
        )


class FusionTransformationScheme(
    Struct, frozen=True, rename={"agency": "agencyId"}
):
    """Fusion-JSON payload for VTL transformation schemes."""

    id: str
    agency: str
    vtlVersion: str
    names: Sequence[FusionString] = ()
    items: Sequence[FusionTransformation] = ()
    vtlMappingScheme: Optional[str] = None
    namePersonalisationScheme: Optional[str] = None
    customTypeScheme: Optional[str] = None
    rulesetSchemes: Sequence[str] = ()
    userDefinedOperatorScheme: Sequence[str] = ()
    descriptions: Sequence[FusionString] = ()
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    isPartial: bool = False

    def to_model(
        self,
        custom_type_schemes: Sequence[FusionCustomTypeScheme],
        vtl_mapping_schemes: Sequence[FusionVtlMappingScheme],
        name_personalisation_schemes: Sequence[
            FusionNamePersonalisationScheme
        ],
        ruleset_schemes: Sequence[FusionRulesetScheme],
        user_defined_operator_schemes: Sequence[
            FusionUserDefinedOperatorScheme
        ],
    ) -> TS:
        """Converts deserialized class to pysdmx model class."""
        cts = [i.to_model() for i in custom_type_schemes]
        vms = [i.to_model() for i in vtl_mapping_schemes]
        nps = [i.to_model() for i in name_personalisation_schemes]
        rss = [i.to_model() for i in ruleset_schemes]
        dos = [i.to_model() for i in user_defined_operator_schemes]
        items = [i.to_model() for i in self.items]
        return TS(
            self.id,
            name=self.names[0].value,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            is_external_reference=self.isExternalReference,
            agency=self.agency,
            items=items,
            is_partial=self.isPartial,
            vtl_version=self.vtlVersion,
            vtl_mapping_scheme=vms[0] if vms else None,
            custom_type_scheme=cts[0] if cts else None,
            name_personalisation_scheme=nps[0] if nps else None,
            ruleset_schemes=rss,
            user_defined_operator_schemes=dos,
        )


class FusionVtlTransformationsMessage(Struct, frozen=True):
    """Fusion-JSON payload for /transformationscheme queries."""

    TransformationScheme: Sequence[FusionTransformationScheme]
    CustomTypeScheme: Sequence[FusionCustomTypeScheme] = ()
    VtlMappingScheme: Sequence[FusionVtlMappingScheme] = ()
    NamePersonalisationScheme: Sequence[FusionNamePersonalisationScheme] = ()
    RulesetScheme: Sequence[FusionRulesetScheme] = ()
    UserDefinedOperatorScheme: Sequence[FusionUserDefinedOperatorScheme] = ()

    def to_model(self) -> TS:
        """Converts deserialized class to pysdmx model class."""
        return self.TransformationScheme[0].to_model(
            self.CustomTypeScheme,
            self.VtlMappingScheme,
            self.NamePersonalisationScheme,
            self.RulesetScheme,
            self.UserDefinedOperatorScheme,
        )



# mypy: disable-error-code="union-attr"
"""Module for writing SDMX-ML 2.1 Structure Specific data messages."""

from typing import Any, Dict, List, Optional, Sequence

import pandas as pd

from pysdmx.io.format import Format
from pysdmx.io.pd import PandasDataset
from pysdmx.io.xml.sdmx21.writer.__write_aux import (
    ABBR_MSG,
    ALL_DIM,
    __escape_xml,
    __write_header,
    create_namespaces,
    get_end_message,
    get_structure,
)
from pysdmx.io.xml.sdmx21.writer.__write_data_aux import (
    check_content_dataset,
    check_dimension_at_observation,
    get_codes,
    writing_validation,
)
from pysdmx.io.xml.sdmx21.writer.config import CHUNKSIZE
from pysdmx.model.message import Header
from pysdmx.util import parse_short_urn


def __memory_optimization_writing(
    dataset: PandasDataset, prettyprint: bool
) -> str:
    """Memory optimization for writing data."""
    outfile = ""
    length_ = len(dataset.data)
    if len(dataset.data) > CHUNKSIZE:
        previous = 0
        next_ = CHUNKSIZE
        while previous <= length_:
            # Sliding a window for efficient access to the data
            # and avoid memory issues
            outfile += __obs_processing(
                dataset.data.iloc[previous:next_], prettyprint
            )
            previous = next_
            next_ += CHUNKSIZE

            if next_ >= length_:
                outfile += __obs_processing(
                    dataset.data.iloc[previous:], prettyprint
                )
                previous = next_
    else:
        outfile += __obs_processing(dataset.data, prettyprint)

    return outfile


def __write_data_structure_specific(
    datasets: Dict[str, PandasDataset],
    dim_mapping: Dict[str, str],
    prettyprint: bool = True,
) -> str:
    """Write data to SDMX-ML 2.1 Structure-Specific format.

    Args:
        datasets: dict. Datasets to be written.
        dim_mapping: dict. URN-DimensionAtObservation mapping.
        prettyprint: bool. Prettyprint or not.

    Returns:
        The data in SDMX-ML 2.1 Structure-Specific format, as string.
    """
    outfile = ""

    for i, (short_urn, dataset) in enumerate(datasets.items()):
        dataset.data = dataset.data.fillna("").astype(str)
        outfile += __write_data_single_dataset(
            dataset=dataset,
            prettyprint=prettyprint,
            count=i + 1,
            dim=dim_mapping[short_urn],
        )

    return outfile


def __write_data_single_dataset(
    dataset: PandasDataset,
    prettyprint: bool = True,
    count: int = 1,
    dim: str = ALL_DIM,
) -> str:
    """Write data to SDMX-ML 2.1 Structure-Specific format.

    Args:
        dataset: PandasDataset. Dataset to be written.
        prettyprint: bool. Prettyprint or not.
        count: int. Count for namespace.
        dim: str. Dimension to be written.

    Returns:
        The data in SDMX-ML 2.1 Structure-Specific format, as string.
    """

    def __remove_optional_attributes_empty_data(str_to_check: str) -> str:
        """This function removes data when optional attributes are found."""
        for att in dataset.structure.components.attributes:
            if not att.required:
                str_to_check = str_to_check.replace(f"{att.id}='' ", "")
                str_to_check = str_to_check.replace(f'{att.id}="" ', "")
        return str_to_check

    outfile = ""
    structure_urn = get_structure(dataset)
    id_structure = parse_short_urn(structure_urn).id
    sdmx_type = parse_short_urn(structure_urn).id

    nl = "\n" if prettyprint else ""
    child1 = "\t" if prettyprint else ""

    attached_attributes_str = ""
    for k, v in dataset.attributes.items():
        attached_attributes_str += f"{k}={str(v)!r} "

    # Datasets
    outfile += (
        f"{nl}{child1}<{ABBR_MSG}:DataSet {attached_attributes_str}"
        f"ss:structureRef={id_structure!r} "
        f'xsi:type="ns{count}:DataSetType" '
        f'ss:dataScope="{sdmx_type}" '
        f'action="{dataset.action.value}">{nl}'
    )
    data = ""
    if dim == ALL_DIM:
        data += __memory_optimization_writing(dataset, prettyprint)
    else:
        writing_validation(dataset)
        series_codes, obs_codes = get_codes(
            dimension_code=dim,
            structure=dataset.structure,  # type: ignore[arg-type]
            data=dataset.data,
        )

        data += __series_processing(
            data=dataset.data,
            series_codes=series_codes,
            obs_codes=obs_codes,
            prettyprint=prettyprint,
        )

        # Remove optional attributes empty data
        data = __remove_optional_attributes_empty_data(data)

    # Adding to outfile
    outfile += data

    outfile += f"{child1}</{ABBR_MSG}:DataSet>"

    return outfile.replace("'", '"')


def __obs_processing(data: pd.DataFrame, prettyprint: bool = True) -> str:
    def __format_obs_str(element: Dict[str, Any]) -> str:
        """Formats the observation as key=value pairs."""
        nl = "\n" if prettyprint else ""
        child2 = "\t\t" if prettyprint else ""

        out = f"{child2}<Obs "

        for k, v in element.items():
            out += f"{k}={__escape_xml(str(v))!r} "

        out += f"/>{nl}"

        return out

    parser = lambda x: __format_obs_str(x)  # noqa: E731

    iterator = map(parser, data.to_dict(orient="records"))

    return "".join(iterator)


def __series_processing(
    data: pd.DataFrame,
    series_codes: List[str],
    obs_codes: List[str],
    prettyprint: bool = True,
) -> str:
    def __generate_series_str() -> str:
        """Generates the series item with its observations."""
        out_list: List[str] = []
        data.groupby(by=series_codes)[obs_codes].apply(
            lambda x: __format_dict_ser(out_list, x)
        )

        return "".join(out_list)

    def __format_dict_ser(
        output_list: List[str],
        obs: Any,
    ) -> Any:
        """Formats the series as key=value pairs."""
        # Creating the observation dict,
        # we always get the first element on Series
        # as we are grouping by it
        data_dict["Series"][0]["Obs"] = obs.to_dict(orient="records")
        output_list.append(__format_ser_str(data_dict["Series"][0]))
        # We remove the data for series as it is no longer necessary
        del data_dict["Series"][0]

    def __format_ser_str(data_info: Dict[Any, Any]) -> str:
        """Formats the series as key=value pairs."""
        child2 = "\t\t" if prettyprint else ""
        child3 = "\t\t\t" if prettyprint else ""
        nl = "\n" if prettyprint else ""

        out_element = f"{child2}<Series "

        for k, v in data_info.items():
            if k != "Obs":
                out_element += f"{k}={__escape_xml(str(v))!r} "

        out_element += f">{nl}"

        for obs in data_info["Obs"]:
            out_element += f"{child3}<Obs "

            for k, v in obs.items():
                out_element += f"{k}={__escape_xml(str(v))!r} "

            out_element += f"/>{nl}"

        out_element += f"{child2}</Series>{nl}"

        return out_element

    # Getting each datapoint from data and creating dict
    data = data.sort_values(series_codes, axis=0)
    data_dict = {
        "Series": data[series_codes]
        .drop_duplicates()
        .reset_index(drop=True)
        .to_dict(orient="records")
    }

    out = __generate_series_str()

    return out


def write(
    datasets: Sequence[PandasDataset],
    output_path: str = "",
    prettyprint: bool = True,
    header: Optional[Header] = None,
    dimension_at_observation: Optional[Dict[str, str]] = None,
) -> Optional[str]:
    """Write data to SDMX-ML 2.1 Structure Specific format.

    Args:
        datasets: The datasets to be written.
        output_path: The path to save the file.
        prettyprint: Prettyprint or not.
        header: The header to be used (generated if None).
        dimension_at_observation:
          The mapping between the dataset and the dimension at observation.

    Returns:
        The XML string if path is empty, None otherwise.
    """
    ss_namespaces = ""
    type_ = Format.DATA_SDMX_ML_2_1_STR

    # Checking if we have datasets,
    # we need to ensure we can write them correctly
    check_content_dataset(datasets)
    content = {dataset.short_urn: dataset for dataset in datasets}

    if header is None:
        header = Header()

    # Checking the dimension at observation mapping
    dim_mapping = check_dimension_at_observation(
        content, dimension_at_observation
    )
    header.structure = dim_mapping
    add_namespace_structure = True
    for i, (short_urn, dimension) in enumerate(header.structure.items()):
        ss_namespaces += (
            f'xmlns:ns{i + 1}="urn:sdmx:org.sdmx'
            f".infomodel.datastructure.{short_urn}"
            f':ObsLevelDim:{dimension}" '
        )

    # Generating the initial tag with namespaces
    outfile = create_namespaces(type_, ss_namespaces, prettyprint)
    # Generating the header
    outfile += __write_header(
        header, prettyprint, add_namespace_structure, data_message=True
    )
    # Writing the content
    outfile += __write_data_structure_specific(
        datasets=content,
        dim_mapping=dim_mapping,
        prettyprint=prettyprint,
    )

    outfile += get_end_message(type_, prettyprint)

    if output_path == "":
        return outfile

    with open(output_path, "w", encoding="UTF-8", errors="replace") as f:
        f.write(outfile)

    return None



"""Collection of Fusion-JSON schemas for categories and category schemes."""

from collections import defaultdict
from typing import Dict, Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.fusion.messages.core import FusionString
from pysdmx.io.json.fusion.messages.dataflow import FusionDataflow
from pysdmx.model import (
    Agency,
    Category,
    DataflowRef,
)
from pysdmx.model import (
    Categorisation as CT,
)
from pysdmx.model import (
    CategoryScheme as CS,
)
from pysdmx.model import (
    Dataflow as DF,
)
from pysdmx.util import find_by_urn


class FusionCategorisation(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a categorisation."""

    id: str
    names: Sequence[FusionString]
    agency: str
    categoryReference: str
    structureReference: str
    descriptions: Optional[Sequence[FusionString]] = None
    version: str = "1.0"

    def to_model(self) -> CT:
        """Converts a JsonCategorisation to a standard categorisation."""
        description = self.descriptions[0].value if self.descriptions else None
        return CT(
            id=self.id,
            name=self.names[0].value,
            agency=self.agency,
            description=description,
            version=self.version,
            source=self.structureReference,
            target=self.categoryReference,
        )


class FusionCategory(Struct, frozen=True):
    """Fusion-JSON payload for a category."""

    id: str
    names: Sequence[FusionString]
    descriptions: Optional[Sequence[FusionString]] = None
    items: Sequence["FusionCategory"] = ()

    def to_model(self) -> Category:
        """Converts a FusionCode to a standard code."""
        description = self.descriptions[0].value if self.descriptions else None
        return Category(
            id=self.id,
            name=self.names[0].value,
            description=description,
            categories=[c.to_model() for c in self.items],
        )


class FusionCategoryScheme(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a category scheme."""

    id: str
    names: Sequence[FusionString]
    agency: str
    descriptions: Optional[Sequence[FusionString]] = None
    version: str = "1.0"
    items: Sequence[FusionCategory] = ()

    def to_model(self) -> CS:
        """Converts a JsonCodelist to a standard codelist."""
        description = self.descriptions[0].value if self.descriptions else None
        return CS(
            id=self.id,
            name=self.names[0].value,
            agency=self.agency,
            description=description,
            version=self.version,
            items=[c.to_model() for c in self.items],
        )


class FusionCategorySchemeMessage(Struct, frozen=True):
    """Fusion-JSON payload for /categoryscheme queries."""

    CategoryScheme: Sequence[FusionCategoryScheme]
    Categorisation: Sequence[FusionCategorisation] = ()
    Dataflow: Sequence[FusionDataflow] = ()

    def __group_flows(self) -> defaultdict[str, list[DF]]:
        out: defaultdict[str, list[DF]] = defaultdict(list)
        for c in self.Categorisation:
            d = find_by_urn(self.Dataflow, c.structureReference)
            src = c.categoryReference[c.categoryReference.find(")") + 2 :]
            out[src].append(d.to_model())
        return out

    def __add_flows(
        self, cat: Category, cni: str, cf: Dict[str, list[DF]]
    ) -> None:
        if cat.categories:
            for c in cat.categories:
                self.__add_flows(c, f"{cni}.{c.id}", cf)
        if cni in cf:
            dfrefs = [
                DataflowRef(
                    (
                        df.agency.id
                        if isinstance(df.agency, Agency)
                        else df.agency
                    ),
                    df.id,
                    df.version,
                    df.name,
                )
                for df in cf[cni]
            ]
            cat.dataflows = dfrefs

    def to_model(self) -> CS:
        """Returns the requested category scheme."""
        cf = self.__group_flows()
        cs = self.CategoryScheme[0].to_model()
        for c in cs:
            self.__add_flows(c, c.id, cf)
        return cs


class FusionCategorisationMessage(Struct, frozen=True):
    """Fusion-JSON payload for /categorisation queries."""

    Categorisation: Sequence[FusionCategorisation]

    def to_model(self) -> Sequence[CT]:
        """Returns the requested categorisations."""
        return [c.to_model() for c in self.Categorisation]



"""Collection of SDMX-JSON schemas for VTL artefacts."""

from datetime import datetime
from typing import Literal, Optional, Sequence

from msgspec import Struct

from pysdmx.model.vtl import (
    CustomType,
    CustomTypeScheme,
    FromVtlMapping,
    NamePersonalisation,
    NamePersonalisationScheme,
    Ruleset,
    RulesetScheme,
    ToVtlMapping,
    Transformation,
    TransformationScheme,
    UserDefinedOperator,
    UserDefinedOperatorScheme,
    VtlCodelistMapping,
    VtlConceptMapping,
    VtlDataflowMapping,
    VtlMapping,
    VtlMappingScheme,
)


class JsonCustomType(Struct, frozen=True):
    """SDMX-JSON payload for custom types."""

    id: str
    name: str
    vtlScalarType: str
    dataType: str
    vtlLiteralFormat: Optional[str] = None
    outputFormat: Optional[str] = None
    nullValue: Optional[str] = None
    description: Optional[str] = None

    def to_model(self) -> CustomType:
        """Converts deserialized class to pysdmx model class."""
        return CustomType(
            id=self.id,
            name=self.name,
            description=self.description,
            data_type=self.dataType,
            null_value=self.nullValue,
            output_format=self.outputFormat,
            vtl_literal_format=self.vtlLiteralFormat,
            vtl_scalar_type=self.vtlScalarType,
        )


class JsonCustomTypeScheme(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for custom type schemes."""

    id: str
    name: str
    agency: str
    vtlVersion: str
    customTypes: Sequence[JsonCustomType] = ()
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    isPartial: bool = False

    def to_model(self) -> CustomTypeScheme:
        """Converts deserialized class to pysdmx model class."""
        items = [i.to_model() for i in self.customTypes]
        return CustomTypeScheme(
            self.id,
            name=self.name,
            description=self.description,
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            is_external_reference=self.isExternalReference,
            agency=self.agency,
            items=items,
            is_partial=self.isPartial,
            vtl_version=self.vtlVersion,
        )


class JsonNamePersonalisation(Struct, frozen=True):
    """SDMX-JSON payload for name personalisations."""

    id: str
    name: str
    vtlDefaultName: str
    personalisedName: str
    vtlArtefact: str
    description: Optional[str] = None

    def to_model(self) -> NamePersonalisation:
        """Converts deserialized class to pysdmx model class."""
        return NamePersonalisation(
            self.id,
            name=self.name,
            description=self.description,
            vtl_default_name=self.vtlDefaultName,
            personalised_name=self.personalisedName,
            vtl_artefact=self.vtlArtefact,
        )


class JsonNamePersonalisationScheme(
    Struct, frozen=True, rename={"agency": "agencyID"}
):
    """SDMX-JSON payload for name personalisation schemes."""

    id: str
    name: str
    agency: str
    vtlVersion: str
    namePersonalisations: Sequence[JsonNamePersonalisation] = ()
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    isPartial: bool = False

    def to_model(self) -> NamePersonalisationScheme:
        """Converts deserialized class to pysdmx model class."""
        items = [i.to_model() for i in self.namePersonalisations]
        return NamePersonalisationScheme(
            self.id,
            name=self.name,
            description=self.description,
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            is_external_reference=self.isExternalReference,
            agency=self.agency,
            items=items,
            is_partial=self.isPartial,
            vtl_version=self.vtlVersion,
        )


class JsonUserDefinedOperator(Struct, frozen=True):
    """SDMX-JSON payload for user defined operator."""

    id: str
    name: str
    operatorDefinition: str
    description: Optional[str] = None

    def to_model(self) -> UserDefinedOperator:
        """Converts deserialized class to pysdmx model class."""
        return UserDefinedOperator(
            self.id,
            name=self.name,
            description=self.description,
            operator_definition=self.operatorDefinition,
        )


class JsonUserDefinedOperatorScheme(
    Struct, frozen=True, rename={"agency": "agencyID"}
):
    """SDMX-JSON payload for user defined operator schemes."""

    id: str
    name: str
    agency: str
    vtlVersion: str
    userDefinedOperators: Sequence[JsonUserDefinedOperator] = ()
    vtlMappingScheme: Optional[str] = None
    rulesetSchemes: Sequence[str] = ()
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    isPartial: bool = False

    def to_model(self) -> UserDefinedOperatorScheme:
        """Converts deserialized class to pysdmx model class."""
        items = [i.to_model() for i in self.userDefinedOperators]
        return UserDefinedOperatorScheme(
            self.id,
            name=self.name,
            description=self.description,
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            is_external_reference=self.isExternalReference,
            agency=self.agency,
            items=items,
            is_partial=self.isPartial,
            vtl_version=self.vtlVersion,
            vtl_mapping_scheme=self.vtlMappingScheme,
            ruleset_schemes=self.rulesetSchemes,
        )


class JsonRuleset(Struct, frozen=True):
    """SDMX-JSON payload for rulesets."""

    id: str
    name: str
    rulesetDefinition: str
    rulesetType: Literal["datapoint", "hierarchical"]
    rulesetScope: Literal["variable", "valuedomain"]
    description: Optional[str] = None

    def to_model(self) -> Ruleset:
        """Converts deserialized class to pysdmx model class."""
        return Ruleset(
            id=self.id,
            name=self.name,
            description=self.description,
            ruleset_definition=self.rulesetDefinition,
            ruleset_type=self.rulesetType,
            ruleset_scope=self.rulesetScope,
        )


class JsonRulesetScheme(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for ruleset schemes."""

    id: str
    name: str
    agency: str
    vtlVersion: str
    rulesets: Sequence[JsonRuleset] = ()
    vtlMappingScheme: Optional[str] = None
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    isPartial: bool = False

    def to_model(self) -> RulesetScheme:
        """Converts deserialized class to pysdmx model class."""
        items = [i.to_model() for i in self.rulesets]
        return RulesetScheme(
            self.id,
            name=self.name,
            description=self.description,
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            is_external_reference=self.isExternalReference,
            agency=self.agency,
            items=items,
            is_partial=self.isPartial,
            vtl_version=self.vtlVersion,
            vtl_mapping_scheme=self.vtlMappingScheme,
        )


class JsonToVtlMapping(Struct, frozen=True):
    """SDMX-JSON payload for To VTL mappings."""

    toVtlSubSpace: Sequence[str]
    type: Optional[str] = None

    def to_model(self) -> ToVtlMapping:
        """Converts deserialized class to pysdmx model class."""
        return ToVtlMapping(self.toVtlSubSpace, self.type)


class JsonFromVtlMapping(Struct, frozen=True):
    """SDMX-JSON payload for from VTL mappings."""

    fromVtlSuperSpace: Sequence[str]
    type: Optional[str] = None

    def to_model(self) -> FromVtlMapping:
        """Converts deserialized class to pysdmx model class."""
        return FromVtlMapping(self.fromVtlSuperSpace, self.type)


class JsonVtlMapping(Struct, frozen=True):
    """SDMX-JSON payload for VTL mappings."""

    id: str
    name: str
    alias: str
    concept: Optional[str] = None
    codelist: Optional[str] = None
    dataflow: Optional[str] = None
    genericDataflow: Optional[str] = None
    description: Optional[str] = None
    toVtlMapping: Optional[JsonToVtlMapping] = None
    fromVtlMapping: Optional[JsonFromVtlMapping] = None

    def to_model(self) -> VtlMapping:
        """Converts deserialized class to pysdmx model class."""
        if self.codelist:
            return VtlCodelistMapping(
                self.id,
                name=self.name,
                description=self.description,
                codelist=self.codelist,
                codelist_alias=self.alias,
            )
        elif self.concept:
            return VtlConceptMapping(
                self.id,
                name=self.name,
                description=self.description,
                concept=self.concept,
                concept_alias=self.alias,
            )
        else:
            return VtlDataflowMapping(
                self.id,
                name=self.name,
                description=self.description,
                dataflow=(
                    self.dataflow if self.dataflow else self.genericDataflow  # type: ignore[arg-type]
                ),
                dataflow_alias=self.alias,
                from_vtl_mapping_method=(
                    self.fromVtlMapping.to_model()
                    if self.fromVtlMapping
                    else None
                ),
                to_vtl_mapping_method=(
                    self.toVtlMapping.to_model() if self.toVtlMapping else None
                ),
            )


class JsonVtlMappingScheme(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for VTL mapping schemes."""

    id: str
    name: str
    agency: str
    vtlMappings: Sequence[JsonVtlMapping] = ()
    vtlMappingScheme: Optional[str] = None
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    isPartial: bool = False

    def to_model(self) -> VtlMappingScheme:
        """Converts deserialized class to pysdmx model class."""
        items = [i.to_model() for i in self.vtlMappings]
        return VtlMappingScheme(
            self.id,
            name=self.name,
            description=self.description,
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            is_external_reference=self.isExternalReference,
            agency=self.agency,
            items=items,
            is_partial=self.isPartial,
        )


class JsonTransformation(Struct, frozen=True):
    """SDMX-JSON payload for VTL transformations."""

    id: str
    name: str
    expression: str
    result: str
    isPersistent: bool
    description: Optional[str] = None

    def to_model(self) -> Transformation:
        """Converts deserialized class to pysdmx model class."""
        return Transformation(
            self.id,
            name=self.name,
            description=self.description,
            expression=self.expression,
            is_persistent=self.isPersistent,
            result=self.result,
        )


class JsonTransformationScheme(
    Struct, frozen=True, rename={"agency": "agencyID"}
):
    """SDMX-JSON payload for VTL transformation schemes."""

    id: str
    name: str
    agency: str
    vtlVersion: str
    transformations: Sequence[JsonTransformation]
    vtlMappingScheme: Optional[str] = None
    namePersonalisationScheme: Optional[str] = None
    customTypeScheme: Optional[str] = None
    rulesetSchemes: Sequence[str] = ()
    userDefinedOperatorSchemes: Sequence[str] = ()
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    isPartial: bool = False

    def to_model(
        self,
        custom_type_schemes: Sequence[JsonCustomTypeScheme],
        vtl_mapping_schemes: Sequence[JsonVtlMappingScheme],
        name_personalisation_schemes: Sequence[JsonNamePersonalisationScheme],
        ruleset_schemes: Sequence[JsonRulesetScheme],
        user_defined_operator_schemes: Sequence[JsonUserDefinedOperatorScheme],
    ) -> TransformationScheme:
        """Converts deserialized class to pysdmx model class."""
        cts = [i.to_model() for i in custom_type_schemes]
        vms = [i.to_model() for i in vtl_mapping_schemes]
        nps = [i.to_model() for i in name_personalisation_schemes]
        rss = [i.to_model() for i in ruleset_schemes]
        dos = [i.to_model() for i in user_defined_operator_schemes]
        items = [i.to_model() for i in self.transformations]
        return TransformationScheme(
            self.id,
            name=self.name,
            description=self.description,
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            is_external_reference=self.isExternalReference,
            agency=self.agency,
            items=items,
            is_partial=self.isPartial,
            vtl_version=self.vtlVersion,
            vtl_mapping_scheme=vms[0] if vms else None,
            custom_type_scheme=cts[0] if cts else None,
            name_personalisation_scheme=nps[0] if nps else None,
            ruleset_schemes=rss,
            user_defined_operator_schemes=dos,
        )


class JsonVtlTransformations(Struct, frozen=True):
    """SDMX-JSON payload for VTL transformation schemes."""

    transformationSchemes: Sequence[JsonTransformationScheme]
    customTypeSchemes: Sequence[JsonCustomTypeScheme] = ()
    vtlMappingSchemes: Sequence[JsonVtlMappingScheme] = ()
    namePersonalisationSchemes: Sequence[JsonNamePersonalisationScheme] = ()
    rulesetSchemes: Sequence[JsonRulesetScheme] = ()
    userDefinedOperatorSchemes: Sequence[JsonUserDefinedOperatorScheme] = ()

    def to_model(self) -> TransformationScheme:
        """Converts deserialized class to pysdmx model class."""
        return self.transformationSchemes[0].to_model(
            self.customTypeSchemes,
            self.vtlMappingSchemes,
            self.namePersonalisationSchemes,
            self.rulesetSchemes,
            self.userDefinedOperatorSchemes,
        )


class JsonVtlTransformationsMessage(Struct, frozen=True):
    """SDMX-JSON payload for /transformationscheme queries."""

    data: JsonVtlTransformations

    def to_model(self) -> TransformationScheme:
        """Returns the requested transformation scheme."""
        return self.data.to_model()



"""Pandas SDMX Dataset."""

from pysdmx.__extras_check import __check_data_extra
from pysdmx.model.dataset import Dataset

__check_data_extra()

# E402 is needed here to ensure a clear message is used on missing import
import pandas as pd  # noqa: E402


class PandasDataset(Dataset, frozen=False, kw_only=True):
    """Class related to Dataset, using Pandas Dataframe.

    It is based on SDMX Dataset and has Pandas Dataframe compatibility to
    withhold data.

    Args:
        attributes: Attributes at dataset level.
        data: Pandas Dataframe.
        structure:
          URN or Schema related to this Dataset
          (DSD, Dataflow, ProvisionAgreement)
    """

    data: pd.DataFrame



"""SDMX 2.1 XML error reader."""

from pysdmx.errors import Invalid
from pysdmx.io.xml.sdmx21.__tokens import (
    ERROR,
    ERROR_CODE,
    ERROR_MESSAGE,
    ERROR_TEXT,
)
from pysdmx.io.xml.sdmx21.reader.__parse_xml import parse_xml


def read(input_str: str, validate: bool = True) -> None:
    """Reads an Error message from the SDMX-ML data and raises the exception.

    Args:
        input_str: The SDMX-ML data as a string.
        validate: If True, the input is validated before

    Raises:
        Invalid: Error message as exception.
    """
    dict_info = parse_xml(input_str, validate=validate)
    if ERROR not in dict_info:
        raise Invalid(
            "This SDMX document is not an SDMX-ML 2.1 Error message."
        )
    code = dict_info[ERROR][ERROR_MESSAGE][ERROR_CODE]
    text = dict_info[ERROR][ERROR_MESSAGE][ERROR_TEXT]
    raise Invalid("Invalid", f"{code}: {text}")



"""Collection of SDMX-JSON schemas for structure map queries."""

from datetime import datetime as dt
from datetime import timezone as tz
from typing import Any, Dict, Literal, Optional, Sequence, Union

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import JsonAnnotation
from pysdmx.model import (
    ComponentMap,
    DataType,
    DatePatternMap,
    FixedValueMap,
    ImplicitComponentMap,
    MultiComponentMap,
    MultiRepresentationMap,
    MultiValueMap,
    RepresentationMap,
    StructureMap,
    ValueMap,
)
from pysdmx.util import find_by_urn


class JsonSourceValue(Struct, frozen=True):
    """SDMX-JSON payload for a source value."""

    value: str
    isRegEx: bool = False

    def to_model(self) -> str:
        """Returns the requested source value."""
        if self.isRegEx:
            return f"regex:{self.value}"
        else:
            return self.value


class JsonRepresentationMapping(Struct, frozen=True):
    """SDMX-JSON payload for a representation mapping."""

    sourceValues: Sequence[JsonSourceValue]
    targetValues: Sequence[str]
    validFrom: Optional[str] = None
    validTo: Optional[str] = None

    def __get_dt(self, inp: str) -> dt:
        return dt.fromisoformat(inp).replace(tzinfo=tz.utc)

    def to_model(self, is_multi: bool) -> Union[MultiValueMap, ValueMap]:
        """Returns the requested value maps."""
        if is_multi:
            return MultiValueMap(
                source=[src.to_model() for src in self.sourceValues],
                target=self.targetValues,
                valid_from=(
                    self.__get_dt(self.validFrom) if self.validFrom else None
                ),
                valid_to=self.__get_dt(self.validTo) if self.validTo else None,
            )
        else:
            return ValueMap(
                source=self.sourceValues[0].to_model(),
                target=self.targetValues[0],
                valid_from=(
                    self.__get_dt(self.validFrom) if self.validFrom else None
                ),
                valid_to=self.__get_dt(self.validTo) if self.validTo else None,
            )


class JsonRepresentationMap(
    Struct,
    frozen=True,
    rename={"agency": "agencyID"},
):
    """SDMX-JSON payload for a representation map."""

    id: str
    name: str
    agency: str
    version: str
    source: Sequence[Dict[str, str]]
    target: Sequence[Dict[str, str]]
    representationMappings: Sequence[JsonRepresentationMapping]
    description: Optional[str] = None
    isExternalReference: bool = False
    validFrom: Optional[dt] = None
    validTo: Optional[dt] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None

    def __parse_st(self, item: Dict[str, str]) -> Union[DataType, str]:
        if "dataType" in item:
            return DataType(item["dataType"])
        elif "valuelist" in item:
            return item["valuelist"]
        else:
            return item["codelist"]

    def to_model(
        self, is_multi: bool = False
    ) -> Union[MultiRepresentationMap, RepresentationMap]:
        """Returns the requested value maps."""
        mrs = [rm.to_model(is_multi) for rm in self.representationMappings]
        s = [self.__parse_st(i) for i in self.source]
        t = [self.__parse_st(j) for j in self.target]
        if is_multi:
            return MultiRepresentationMap(
                id=self.id,
                name=self.name,
                agency=self.agency,
                source=s,
                target=t,
                maps=mrs,  # type: ignore[arg-type]
                description=self.description,
                version=self.version,
            )
        else:
            return RepresentationMap(
                id=self.id,
                name=self.name,
                agency=self.agency,
                source=s[0],
                target=t[0],
                maps=mrs,  # type: ignore[arg-type]
                description=self.description,
                version=self.version,
            )


class JsonFixedValueMap(Struct, frozen=True):
    """SDMX-JSON payload for a fixed value map."""

    values: Sequence[Any]
    source: Optional[str] = None
    target: Optional[str] = None

    def to_model(self) -> FixedValueMap:
        """Returns the requested fixed value map."""
        located_in = "source" if self.source else "target"
        target = self.target if self.target else self.source
        return FixedValueMap(
            target,  # type: ignore[arg-type]
            self.values[0],
            located_in,  # type: ignore[arg-type]
        )


class JsonComponentMap(Struct, frozen=True):
    """SDMX-JSON payload for a component map."""

    source: Sequence[str]
    target: Sequence[str]
    representationMap: Optional[str] = None

    def to_model(
        self, rms: Sequence[JsonRepresentationMap]
    ) -> Union[ComponentMap, MultiComponentMap, ImplicitComponentMap]:
        """Returns the requested map."""
        if self.representationMap:
            rm = find_by_urn(rms, self.representationMap)
            if len(self.source) == 1 and len(self.target) == 1:
                return ComponentMap(
                    self.source[0],
                    self.target[0],
                    rm.to_model(),
                )
            else:
                return MultiComponentMap(
                    self.source, self.target, rm.to_model(True)
                )
        else:
            return ImplicitComponentMap(self.source[0], self.target[0])


class JsonMappedPair(Struct, frozen=True):
    """SDMX-JSON payload for a pair of mapped components."""

    source: str
    target: str


class JsonDatePatternMap(Struct, frozen=True):
    """SDMX-JSON payload for a date pattern map."""

    sourcePattern: str
    mappedComponents: Sequence[JsonMappedPair]
    locale: str
    id: Optional[str] = None
    resolvePeriod: Optional[
        Literal["startOfPeriod", "endOfPeriod", "midPeriod"]
    ] = None
    targetFrequencyID: Optional[str] = None
    frequencyDimension: Optional[str] = None
    mappedFrequencies: Optional[Sequence[str]] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None

    def to_model(self) -> DatePatternMap:
        """Returns the requested date mapper."""
        freq = (
            self.targetFrequencyID
            if self.targetFrequencyID
            else self.frequencyDimension
        )
        typ = "fixed" if self.targetFrequencyID else "variable"
        return DatePatternMap(
            self.mappedComponents[0].source,
            self.mappedComponents[0].target,
            self.sourcePattern,
            freq,  # type: ignore[arg-type]
            self.id,
            self.locale,
            typ,  # type: ignore[arg-type]
        )


class JsonStructureMap(Struct, frozen=True):
    """SDMX-JSON payload for a structure map."""

    id: str
    name: str
    version: str
    agencyID: str
    source: str
    target: str
    description: Optional[str] = None
    isExternalReference: bool = False
    validFrom: Optional[dt] = None
    validTo: Optional[dt] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    datePatternMaps: Sequence[JsonDatePatternMap] = ()
    componentMaps: Sequence[JsonComponentMap] = ()
    fixedValueMaps: Sequence[JsonFixedValueMap] = ()

    def to_model(
        self,
        rms: Sequence[JsonRepresentationMap],
    ) -> StructureMap:
        """Returns the requested mapping definition."""
        m1 = tuple([dpm.to_model() for dpm in self.datePatternMaps])
        m2 = tuple([cm.to_model(rms) for cm in self.componentMaps])
        m3 = tuple([fvm.to_model() for fvm in self.fixedValueMaps])
        return StructureMap(
            id=self.id,
            name=self.name,
            agency=self.agencyID,
            source=self.source,
            target=self.target,
            maps=m1 + m2 + m3,
            description=self.description,
            version=self.version,
        )


class JsonStructureMaps(Struct, frozen=True):
    """SDMX-JSON payload for structure maps."""

    structureMaps: Sequence[JsonStructureMap]
    representationMaps: Sequence[JsonRepresentationMap] = ()

    def to_model(self) -> StructureMap:
        """Returns the requested mapping definition."""
        return self.structureMaps[0].to_model(self.representationMaps)


class JsonMappingMessage(Struct, frozen=True):
    """SDMX-JSON payload for /structuremap queries."""

    data: JsonStructureMaps

    def to_model(self) -> StructureMap:
        """Returns the requested mapping definition."""
        return self.data.to_model()


class JsonRepresentationMaps(Struct, frozen=True):
    """SDMX-JSON payload for representation maps."""

    representationMaps: Sequence[JsonRepresentationMap]

    def to_model(self) -> Sequence[ValueMap]:
        """Returns the requested mapping definition."""
        m = self.representationMaps[0]
        multi = bool(len(m.source) > 1 or len(m.target) > 1)
        out = m.to_model(multi)
        return out  # type: ignore[return-value]


class JsonRepresentationMapMessage(Struct, frozen=True):
    """SDMX-JSON payload for /representationmap queries."""

    data: JsonRepresentationMaps

    def to_model(self) -> Sequence[ValueMap]:
        """Returns the requested representation map."""
        return self.data.to_model()



"""Collection of SDMX-JSON schemas for common artefacts."""

from datetime import datetime
from typing import Optional, Sequence, Union

import msgspec

from pysdmx.errors import NotFound
from pysdmx.model import ArrayBoundaries, Codelist, Facets
from pysdmx.util import find_by_urn


class JsonAnnotation(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for annotations."""

    id: Optional[str] = None
    title: Optional[str] = None
    type: Optional[str] = None
    value: Optional[str] = None
    text: Optional[str] = None


class JsonTextFormat(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for TextFormat."""

    dataType: str
    minLength: Optional[int] = None
    maxLength: Optional[int] = None
    minValue: Optional[Union[int, float]] = None
    maxValue: Optional[Union[int, float]] = None
    startValue: Optional[Union[int, float]] = None
    endValue: Optional[Union[int, float]] = None
    decimals: Optional[int] = None
    pattern: Optional[str] = None
    startTime: Optional[datetime] = None
    endTime: Optional[datetime] = None
    isSequence: bool = False
    isMultilingual: bool = False
    sentinelValues: Optional[Sequence[str]] = None
    timeInterval: Optional[str] = None
    interval: Optional[int] = None


class JsonRepresentation(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for core representation."""

    enumerationFormat: Optional[JsonTextFormat] = None
    enumeration: Optional[str] = None
    format: Optional[JsonTextFormat] = None
    minOccurs: Optional[int] = None
    maxOccurs: Optional[int] = None

    def to_facets(self) -> Optional[Facets]:
        """Return a Facets domain object."""
        fmt = self.enumerationFormat if self.enumeration else self.format
        if fmt and (
            fmt.minLength
            or fmt.maxLength
            or fmt.isSequence
            or fmt.minValue
            or fmt.maxValue
            or fmt.startValue
            or fmt.endValue
            or fmt.decimals
            or fmt.pattern
            or fmt.startTime
            or fmt.endTime
        ):
            return Facets(
                min_length=fmt.minLength,
                max_length=fmt.maxLength,
                is_sequence=fmt.isSequence,
                min_value=fmt.minValue,
                max_value=fmt.maxValue,
                start_value=fmt.startValue,
                end_value=fmt.endValue,
                decimals=fmt.decimals,
                pattern=fmt.pattern,
                start_time=fmt.startTime,
                end_time=fmt.endTime,
            )
        else:
            return None

    def to_enumeration(
        self,
        codelists: Sequence[Codelist],
        valid: Sequence[str],
    ) -> Optional[Codelist]:
        """Returns the list of codes allowed for this component."""
        if self.enumeration:
            try:
                a = find_by_urn(codelists, self.enumeration)
                codes = [c for c in a.codes if not valid or c.id in valid]
                return msgspec.structs.replace(a, items=codes)
            except NotFound:
                # This is OK. In case of schema queries, if a component
                # has both a local and a core representations, only the
                # relevant one (i.e. the local one) will be included in
                # the schema.
                return None
        return None

    def to_array_def(self) -> Optional[ArrayBoundaries]:
        """Returns the array boundaries, if any."""
        if self.maxOccurs and self.maxOccurs > 1:
            m = self.minOccurs if self.minOccurs is not None else 0
            return ArrayBoundaries(m, self.maxOccurs)
        else:
            return None


class JsonLink(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for link objects."""

    urn: str
    rel: Optional[str] = None


class JsonHeader(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for message header."""

    links: Sequence[JsonLink] = ()



from parsy import (  # type: ignore[import-untyped]
    ParseError,
    alt,
    any_char,
    regex,
    string,
)

from pysdmx.errors import Invalid

__YEAR4 = string("yyyy").map(lambda x: r"%Y")
__YEAR2 = string("yy").map(lambda x: r"%y")
__MONTH_F = string("MMMM").map(lambda x: r"%B")
__MONTH_S = string("MMM").map(lambda x: r"%b")
__MONTH_NUM = string("MM").map(lambda x: r"%m")
__WEEK_Y = string("ww").map(lambda x: r"%V")
__DAY_Y = string("DD").map(lambda x: r"%j")
__DAY_M = string("dd").map(lambda x: r"%d")
__DAY_NUM = string("U").map(lambda x: r"%u")
__DAY_NM_F = string("EEEE").map(lambda x: r"%A")
__DAY_NM_S = regex("^E{1,3}$").map(lambda x: r"%a")
__HOUR_DAY = string("HH").map(lambda x: r"%H")
__HOUR_AP = string("hh").map(lambda x: r"%I")
__MIN = string("mm").map(lambda x: r"%M")
__SEC = string("ss").map(lambda x: r"%S")
__ANY = any_char

__single_parser = alt(
    __YEAR4,
    __YEAR2,
    __MONTH_F,
    __MONTH_S,
    __MONTH_NUM,
    __WEEK_Y,
    __DAY_Y,
    __DAY_M,
    __DAY_NUM,
    __DAY_NM_F,
    __DAY_NM_S,
    __HOUR_DAY,
    __HOUR_AP,
    __MIN,
    __SEC,
    __ANY,
)

__dpm_parser = __single_parser.at_least(1)


def convert_dpm(sdmx_pattern: str) -> str:
    """Convert an SDMX date pattern into Python format codes."""
    unsupported = ["G", "n", "kk", "KK", "S", "W"]
    for i in unsupported:
        if i in sdmx_pattern:
            raise Invalid("Parsing failed", f"Pattern {i} is not supported.")
    try:
        return "".join(__dpm_parser.parse(sdmx_pattern))
    except ParseError as pe:
        raise Invalid("Invalid pattern", f"The error was: {str(pe)}") from pe



"""Build SDMX-REST structure queries."""

from datetime import datetime
from enum import Enum
from typing import Optional, Sequence, Union

import msgspec

from pysdmx.api.qb.util import (
    REST_ALL,
    REST_LATEST,
    ApiVersion,
    CoreQuery,
    check_multiple_items,
)
from pysdmx.errors import Invalid
from pysdmx.io.format import StructureFormat


class StructureDetail(Enum):
    """The desired amount of information to be returned."""

    FULL = "full"
    ALL_STUBS = "allstubs"
    REFERENCE_STUBS = "referencestubs"
    ALL_COMPLETE_STUBS = "allcompletestubs"
    REFERENCE_COMPLETE_STUBS = "referencecompletestubs"
    REFERENCE_PARTIAL = "referencepartial"
    RAW = "raw"


class StructureReference(Enum):
    """Which additional artefacts to include in the response."""

    NONE = "none"
    PARENTS = "parents"
    PARENTSANDSIBLINGS = "parentsandsiblings"
    ANCESTORS = "ancestors"
    CHILDREN = "children"
    DESCENDANTS = "descendants"
    ALL = "all"
    DATA_STRUCTURE = "datastructure"
    METADATA_STRUCTURE = "metadatastructure"
    CATEGORY_SCHEME = "categoryscheme"
    CONCEPT_SCHEME = "conceptscheme"
    CODELIST = "codelist"
    HIERARCHICAL_CODELIST = "hierarchicalcodelist"
    ORGANISATION_SCHEME = "organisationscheme"
    AGENCY_SCHEME = "agencyscheme"
    DATA_PROVIDER_SCHEME = "dataproviderscheme"
    DATA_CONSUMER_SCHEME = "dataconsumerscheme"
    ORGANISATION_UNIT_SCHEME = "organisationunitscheme"
    DATAFLOW = "dataflow"
    METADATAFLOW = "metadataflow"
    REPORTING_TAXONOMY = "reportingtaxonomy"
    PROVISION_AGREEMENT = "provisionagreement"
    STRUCTURE_SET = "structureset"
    PROCESS = "process"
    CATEGORISATION = "categorisation"
    CONTENT_CONSTRAINT = "contentconstraint"
    ATTACHMENT_CONSTRAINT = "attachmentconstraint"
    ACTUAL_CONSTRAINT = "actualconstraint"
    ALLOWED_CONSTRAINT = "allowedconstraint"
    TRANSFORMATION_SCHEME = "transformationscheme"
    RULESET_SCHEME = "rulesetscheme"
    USER_DEFINED_OPERATOR_SCHEME = "userdefinedoperatorscheme"
    CUSTOM_TYPE_SCHEME = "customtypescheme"
    NAME_PERSONALISATION_SCHEME = "namepersonalisationscheme"
    NAME_ALIAS_SCHEME = "namealiasscheme"
    DATA_CONSTRAINT = "dataconstraint"
    METADATA_CONSTRAINT = "metadataconstraint"
    HIERARCHY = "hierarchy"
    HIERARCHY_ASSOCIATION = "hierarchyassociation"
    VTL_MAPPING_SCHEME = "vtlmappingscheme"
    VALUE_LIST = "valuelist"
    STRUCTURE_MAP = "structuremap"
    REPRESENTATION_MAP = "representationmap"
    CONCEPT_SCHEME_MAP = "conceptschememap"
    CATEGORY_SCHEME_MAP = "categoryschememap"
    ORGANISATION_SCHEME_MAP = "organisationschememap"
    REPORTING_TAXONOMY_MAP = "reportingtaxonomymap"
    METADATA_PROVIDER_SCHEME = "metadataproviderscheme"
    METADATA_PROVISION_AGREEMENT = "metadataprovisionagreement"

    def is_artefact_type(self) -> bool:
        """Whether the references points to a type of artefacts."""
        core_refs = [
            "none",
            "parents",
            "parentsandsiblings",
            "ancestors",
            "children",
            "descendants",
            "all",
        ]
        return self.value not in core_refs


class StructureType(Enum):
    """The type of structural metadata to be returned."""

    DATA_STRUCTURE = "datastructure"
    METADATA_STRUCTURE = "metadatastructure"
    CATEGORY_SCHEME = "categoryscheme"
    CONCEPT_SCHEME = "conceptscheme"
    CODELIST = "codelist"
    HIERARCHICAL_CODELIST = "hierarchicalcodelist"
    ORGANISATION_SCHEME = "organisationscheme"
    AGENCY_SCHEME = "agencyscheme"
    DATA_PROVIDER_SCHEME = "dataproviderscheme"
    DATA_CONSUMER_SCHEME = "dataconsumerscheme"
    ORGANISATION_UNIT_SCHEME = "organisationunitscheme"
    DATAFLOW = "dataflow"
    METADATAFLOW = "metadataflow"
    REPORTING_TAXONOMY = "reportingtaxonomy"
    PROVISION_AGREEMENT = "provisionagreement"
    STRUCTURE_SET = "structureset"
    PROCESS = "process"
    CATEGORISATION = "categorisation"
    CONTENT_CONSTRAINT = "contentconstraint"
    ATTACHMENT_CONSTRAINT = "attachmentconstraint"
    ACTUAL_CONSTRAINT = "actualconstraint"
    ALLOWED_CONSTRAINT = "allowedconstraint"
    TRANSFORMATION_SCHEME = "transformationscheme"
    RULESET_SCHEME = "rulesetscheme"
    USER_DEFINED_OPERATOR_SCHEME = "userdefinedoperatorscheme"
    CUSTOM_TYPE_SCHEME = "customtypescheme"
    NAME_PERSONALISATION_SCHEME = "namepersonalisationscheme"
    NAME_ALIAS_SCHEME = "namealiasscheme"
    DATA_CONSTRAINT = "dataconstraint"
    METADATA_CONSTRAINT = "metadataconstraint"
    HIERARCHY = "hierarchy"
    HIERARCHY_ASSOCIATION = "hierarchyassociation"
    VTL_MAPPING_SCHEME = "vtlmappingscheme"
    VALUE_LIST = "valuelist"
    STRUCTURE_MAP = "structuremap"
    REPRESENTATION_MAP = "representationmap"
    CONCEPT_SCHEME_MAP = "conceptschememap"
    CATEGORY_SCHEME_MAP = "categoryschememap"
    ORGANISATION_SCHEME_MAP = "organisationschememap"
    REPORTING_TAXONOMY_MAP = "reportingtaxonomymap"
    METADATA_PROVIDER_SCHEME = "metadataproviderscheme"
    METADATA_PROVISION_AGREEMENT = "metadataprovisionagreement"
    ALL = REST_ALL


ITEM_SCHEMES = {
    StructureType.CATEGORY_SCHEME,
    StructureType.CONCEPT_SCHEME,
    StructureType.CODELIST,
    StructureType.ORGANISATION_SCHEME,
    StructureType.AGENCY_SCHEME,
    StructureType.DATA_PROVIDER_SCHEME,
    StructureType.METADATA_PROVIDER_SCHEME,
    StructureType.DATA_CONSUMER_SCHEME,
    StructureType.ORGANISATION_UNIT_SCHEME,
    StructureType.REPORTING_TAXONOMY,
    StructureType.TRANSFORMATION_SCHEME,
    StructureType.RULESET_SCHEME,
    StructureType.USER_DEFINED_OPERATOR_SCHEME,
    StructureType.CUSTOM_TYPE_SCHEME,
    StructureType.NAME_PERSONALISATION_SCHEME,
    StructureType.NAME_ALIAS_SCHEME,
    StructureType.VTL_MAPPING_SCHEME,
    StructureType.VALUE_LIST,
    StructureType.HIERARCHICAL_CODELIST,
}

_INITIAL_RESOURCES = {
    StructureType.DATA_STRUCTURE,
    StructureType.METADATA_STRUCTURE,
    StructureType.CATEGORY_SCHEME,
    StructureType.CONCEPT_SCHEME,
    StructureType.CODELIST,
    StructureType.HIERARCHICAL_CODELIST,
    StructureType.ORGANISATION_SCHEME,
    StructureType.AGENCY_SCHEME,
    StructureType.DATA_PROVIDER_SCHEME,
    StructureType.DATA_CONSUMER_SCHEME,
    StructureType.ORGANISATION_UNIT_SCHEME,
    StructureType.DATAFLOW,
    StructureType.METADATAFLOW,
    StructureType.REPORTING_TAXONOMY,
    StructureType.PROVISION_AGREEMENT,
    StructureType.STRUCTURE_SET,
    StructureType.PROCESS,
    StructureType.CATEGORISATION,
    StructureType.CONTENT_CONSTRAINT,
    StructureType.ATTACHMENT_CONSTRAINT,
    StructureType.ALL,
}

_V1_3_RESOURCES = {
    StructureType.ACTUAL_CONSTRAINT,
    StructureType.ALLOWED_CONSTRAINT,
}.union(_INITIAL_RESOURCES)

_V1_5_RESOURCES = {
    StructureType.TRANSFORMATION_SCHEME,
    StructureType.RULESET_SCHEME,
    StructureType.USER_DEFINED_OPERATOR_SCHEME,
    StructureType.CUSTOM_TYPE_SCHEME,
    StructureType.NAME_PERSONALISATION_SCHEME,
    StructureType.NAME_ALIAS_SCHEME,
}.union(_V1_3_RESOURCES)

_V2_0_DEPRECATED = {
    StructureType.HIERARCHICAL_CODELIST,
    StructureType.ORGANISATION_SCHEME,
    StructureType.STRUCTURE_SET,
    StructureType.CONTENT_CONSTRAINT,
    StructureType.ATTACHMENT_CONSTRAINT,
    StructureType.ACTUAL_CONSTRAINT,
    StructureType.ALLOWED_CONSTRAINT,
    StructureType.NAME_ALIAS_SCHEME,
}

_V2_0_ADDED = {
    StructureType.DATA_CONSTRAINT,
    StructureType.METADATA_CONSTRAINT,
    StructureType.HIERARCHY,
    StructureType.HIERARCHY_ASSOCIATION,
    StructureType.VTL_MAPPING_SCHEME,
    StructureType.VALUE_LIST,
    StructureType.STRUCTURE_MAP,
    StructureType.REPRESENTATION_MAP,
    StructureType.CONCEPT_SCHEME_MAP,
    StructureType.CATEGORY_SCHEME_MAP,
    StructureType.ORGANISATION_SCHEME_MAP,
    StructureType.REPORTING_TAXONOMY_MAP,
    StructureType.METADATA_PROVIDER_SCHEME,
    StructureType.METADATA_PROVISION_AGREEMENT,
}

_V2_0_RESOURCES = (_V1_5_RESOURCES.difference(_V2_0_DEPRECATED)).union(
    _V2_0_ADDED
)

_API_RESOURCES = {
    "V1.0.0": _INITIAL_RESOURCES,
    "V1.0.1": _INITIAL_RESOURCES,
    "V1.0.2": _INITIAL_RESOURCES,
    "V1.1.0": _INITIAL_RESOURCES,
    "V1.2.0": _INITIAL_RESOURCES,
    "V1.3.0": _V1_3_RESOURCES,
    "V1.4.0": _V1_3_RESOURCES,
    "V1.5.0": _V1_5_RESOURCES,
    "V2.0.0": _V2_0_RESOURCES,
    "V2.1.0": _V2_0_RESOURCES,
    "V2.2.0": _V2_0_RESOURCES,
    "LATEST": _V2_0_RESOURCES,
}


class StructureQuery(CoreQuery, frozen=True, omit_defaults=True):
    """A query for structural metadata.

    Attributes:
        artefact_type: The type of structural metadata to be returned.
        agency_id: The agency (or agencies) maintaining the artefact(s)
            to be returned.
        resource_id: The id(s) of the artefact(s) to be returned.
        version: The version(s) of the artefact(s) to be returned.
        item_id: The id(s) of the item(s) to be returned.
        detail: The desired amount of information to be returned.
        references: The additional artefact(s) to include in the response.
        as_of: Retrieve the artefact as it was at the specified point
            in time (aka time travel).
    """

    artefact_type: StructureType = StructureType.ALL
    agency_id: Union[str, Sequence[str]] = REST_ALL
    resource_id: Union[str, Sequence[str]] = REST_ALL
    version: Union[str, Sequence[str]] = REST_LATEST
    item_id: Union[str, Sequence[str]] = REST_ALL
    detail: StructureDetail = StructureDetail.FULL
    references: StructureReference = StructureReference.NONE
    as_of: Optional[datetime] = None

    def _validate_query(self, version: ApiVersion) -> None:
        self.validate()
        self.__check_multiple_items(version)
        self.__check_artefact_type(self.artefact_type, version)
        self.__check_item(version)
        self.__check_detail(version)
        self.__check_references(version)
        self.__check_as_of(version)

    def __check_multiple_items(self, version: ApiVersion) -> None:
        check_multiple_items(self.agency_id, version)
        check_multiple_items(self.resource_id, version)
        check_multiple_items(self.version, version)
        check_multiple_items(self.item_id, version)

    def __check_as_of(self, version: ApiVersion) -> None:
        if self.as_of and version < ApiVersion.V2_2_0:
            raise Invalid(
                "Validation Error",
                f"as_of not supported in {version.value}.",
            )

    def __check_artefact_type(
        self, atyp: StructureType, version: ApiVersion
    ) -> None:
        if atyp not in _API_RESOURCES[version.name.replace("_", ".")]:
            raise Invalid(
                "Validation Error",
                f"{atyp} is not valid for SDMX-REST {version.name}.",
            )

    def __check_item(self, version: ApiVersion) -> None:
        if self.item_id != REST_ALL and version < ApiVersion.V1_1_0:
            raise Invalid(
                "Validation Error",
                f"Item query not supported in {version.value}.",
            )

    def __check_detail(self, version: ApiVersion) -> None:
        if (
            version < ApiVersion.V1_3_0
            and self.detail.value
            in [
                "referencepartial",
                "allcompletestubs",
                "referencecompletestubs",
            ]
        ) or (version < ApiVersion.V2_0_0 and self.detail.value == "raw"):
            raise Invalid(
                "Validation Error",
                f"{self.detail} not allowed in SDMX-REST {version.value}.",
            )

    def __check_references(self, version: ApiVersion) -> None:
        if self.references.is_artefact_type():
            rt = StructureType(self.references.value)
            self.__check_artefact_type(rt, version)
        elif (
            self.references == StructureReference.ANCESTORS
            and version < ApiVersion.V2_0_0
        ):
            raise Invalid(
                "Validation Error",
                f"{self.references} not allowed in SDMX-REST {version.value}.",
            )

    def __is_item_allowed(
        self, typ: StructureType, version: ApiVersion
    ) -> bool:
        if typ in ITEM_SCHEMES and version > ApiVersion.V1_0_2:
            return not (
                version == ApiVersion.V1_1_0
                and typ == StructureType.HIERARCHICAL_CODELIST
            )
        else:
            return False

    def __to_type_kw(self, val: StructureType, ver: ApiVersion) -> str:
        if val == StructureType.ALL and ver < ApiVersion.V2_0_0:
            out = "structure"
        else:
            out = val.value
        return out

    def __to_kws(
        self, vals: Union[str, Sequence[str]], ver: ApiVersion
    ) -> str:
        vals = [vals] if isinstance(vals, str) else vals
        mapped = [self._to_kw(v, ver) for v in vals]
        sep = "+" if ver < ApiVersion.V2_0_0 else ","
        return sep.join(mapped)

    def _create_full_query(self, ver: ApiVersion) -> str:
        u = "/"
        u += "structure/" if ver >= ApiVersion.V2_0_0 else ""
        t = self.__to_type_kw(self.artefact_type, ver)
        a = self.__to_kws(self.agency_id, ver)
        r = self.__to_kws(self.resource_id, ver)
        v = self.__to_kws(self.version, ver)
        i = self.__to_kws(self.item_id, ver)
        u += f"{t}/{a}/{r}/{v}"
        ck = [self.__is_item_allowed(self.artefact_type, ver)]
        u += f"/{i}" if all(ck) else ""
        u += f"?detail={self.detail.value}&references={self.references.value}"
        if self.as_of:
            u += f'&asOf={self.as_of.isoformat("T", "seconds")}'
        return u

    def _create_short_query(self, ver: ApiVersion) -> str:
        u = "/structure" if ver >= ApiVersion.V2_0_0 else ""
        t = self.__to_type_kw(self.artefact_type, ver)
        a = self.__to_kws(self.agency_id, ver)
        r = self.__to_kws(self.resource_id, ver)
        v = self.__to_kws(self.version, ver)
        i = self.__to_kws(self.item_id, ver)

        ck = self.__is_item_allowed(self.artefact_type, ver)
        iu = f"/{i}" if ck and self.item_id != REST_ALL else ""
        vu = f"/{v}{iu}" if iu or self.version != REST_LATEST else ""
        ru = f"/{r}{vu}" if vu or self.resource_id != REST_ALL else ""
        au = f"/{a}{ru}" if ru or self.agency_id != REST_ALL else ""

        if ver >= ApiVersion.V2_0_0:
            tu = (
                f"/{t}{au}"
                if au or self.artefact_type != StructureType.ALL
                else ""
            )
        else:
            tu = f"/{t}{au}"
        u += f"{tu}"
        u += (
            "?"
            if self.detail != StructureDetail.FULL
            or self.references != StructureReference.NONE
            or self.as_of
            else ""
        )
        u += (
            f"detail={self.detail.value}"
            if self.detail != StructureDetail.FULL
            else ""
        )
        u += (
            "&"
            if self.detail != StructureDetail.FULL
            and self.references != StructureReference.NONE
            else ""
        )
        u += (
            f"references={self.references.value}"
            if self.references != StructureReference.NONE
            else ""
        )
        u += (
            "&"
            if self.as_of
            and (
                self.detail != StructureDetail.FULL
                or self.references != StructureReference.NONE
            )
            else ""
        )
        u += (
            f'asOf={self.as_of.isoformat("T", "seconds")}'
            if self.as_of
            else ""
        )
        return u

    def _get_decoder(self) -> msgspec.json.Decoder:  # type: ignore[type-arg]
        return _decoder


_decoder = msgspec.json.Decoder(StructureQuery)


__all__ = [
    "ApiVersion",
    "StructureDetail",
    "StructureFormat",
    "StructureQuery",
    "StructureReference",
    "StructureType",
]



from typing import Any, Dict

import xmltodict

from pysdmx.io.xml.sdmx21.reader.doc_validation import validate_doc

SCHEMA_ROOT = "http://www.sdmx.org/resources/sdmxml/schemas/v2_1/"
NAMESPACES_21 = {
    SCHEMA_ROOT + "message": None,
    SCHEMA_ROOT + "common": None,
    SCHEMA_ROOT + "structure": None,
    "http://www.w3.org/2001/XMLSchema-instance": "xsi",
    "http://www.w3.org/XML/1998/namespace": None,
    SCHEMA_ROOT + "data/structurespecific": None,
    SCHEMA_ROOT + "data/generic": None,
    SCHEMA_ROOT + "registry": None,
    "http://schemas.xmlsoap.org/soap/envelope/": None,
}

XML_OPTIONS = {
    "process_namespaces": True,
    "namespaces": NAMESPACES_21,
    "dict_constructor": dict,
    "attr_prefix": "",
}


def parse_xml(
    input_str: str,
    validate: bool = True,
) -> Dict[str, Any]:
    """Reads SDMX-ML data and returns a dictionary with the parsed data.

    Args:
        input_str: SDMX-ML data to be parsed.
        validate: If True, the SDMX-ML data will be validated against the XSD.

    Returns:
        dict: Dictionary with the parsed data.

    Raises:
        Invalid: If the SDMX data cannot be parsed.
    """
    if validate:
        validate_doc(input_str)
    dict_info = xmltodict.parse(
        input_str,
        **XML_OPTIONS,  # type: ignore[arg-type]
    )

    del input_str

    return dict_info



"""Collection of Fusion-JSON schemas for concepts."""

from typing import Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.fusion.messages.code import FusionCodelist
from pysdmx.io.json.fusion.messages.core import (
    FusionRepresentation,
    FusionString,
)
from pysdmx.model.concept import Concept, DataType
from pysdmx.model.concept import ConceptScheme as CS


class FusionConcept(Struct, frozen=True):
    """Fusion-JSON payload for concepts."""

    id: str
    names: Sequence[FusionString]
    representation: Optional[FusionRepresentation] = None
    descriptions: Optional[Sequence[FusionString]] = None

    def to_model(self, codelists: Sequence[FusionCodelist]) -> Concept:
        """Converts a FusionConcept to a standard concept."""
        dt = (
            DataType(self.representation.textFormat.textType)
            if self.representation and self.representation.textFormat
            else DataType.STRING
        )
        f = self.representation.to_facets() if self.representation else None
        c = (
            self.representation.to_enumeration(codelists, [])
            if self.representation
            else None
        )
        d = self.descriptions[0].value if self.descriptions else None
        cl_ref = (
            self.representation.representation
            if self.representation and c
            else None
        )
        return Concept(
            id=self.id,
            dtype=dt,
            facets=f,
            name=self.names[0].value,
            description=d,
            codes=c,
            enum_ref=cl_ref,
        )


class FusionConceptScheme(Struct, frozen=True, rename={"agency": "agencyId"}):
    """Fusion-JSON payload for a concept scheme."""

    id: str
    names: Sequence[FusionString]
    agency: str
    descriptions: Optional[Sequence[FusionString]] = None
    version: str = "1.0"
    items: Sequence[FusionConcept] = ()

    def to_model(self, codelists: Sequence[FusionCodelist]) -> CS:
        """Converts a FusionConceptScheme to a standard concept scheme."""
        d = self.descriptions[0].value if self.descriptions else None
        return CS(
            id=self.id,
            name=self.names[0].value,
            agency=self.agency,
            description=d,
            version=self.version,
            items=[c.to_model(codelists) for c in self.items],
        )


class FusionConceptSchemeMessage(
    Struct,
    frozen=True,
):
    """Fusion-JSON payload for /conceptscheme queries."""

    ConceptScheme: Sequence[FusionConceptScheme]
    Codelist: Sequence[FusionCodelist] = ()

    def to_model(self) -> CS:
        """Returns the requested concept scheme."""
        return self.ConceptScheme[0].to_model(self.Codelist)
